#
# -*- coding: utf-8 -*-


############################################################################################################
#
#   YELLOW PAGES
#
#===========================================================================================================
#   from __future__ import division
#-----------------------------------------------------------------------------------------------------------
EzClassMc                               = g.configtool.utilities.EzClassMc.EzClassMc
#-----------------------------------------------------------------------------------------------------------


############################################################################################################
#
#
#
#===========================================================================================================
class DsatDescription:
    """Class to hold a simplified description of a data source attribute ('dsat'). Used by
    ``PsfCmfIdioms``."""

    #-------------------------------------------------------------------------------------------------------
    __metaclass__   = EzClassMc
    __slots__       = True
    __obligatory__  = """id tablealias attributename""".split()
    __optional__    = dict(
        fktargetalias       = '',
        dbType              = 'text',
        nosearch            = 0,        # if set to 1 no search is allowed on this attribute
        rsindex             = None,     # only needed for retrieving data from record sets
        sortentry           = None,
        dottedLocator       = None,
        )

    #-------------------------------------------------------------------------------------------------------
    def getSortentry( self ):
        return ( self.tablealias, self.fktargetalias, self.attributename )

    #-------------------------------------------------------------------------------------------------------
    def asDBAttributeObject( self ):
        from biocase.wrapper.typeconverter import DBAttributeClass
        R = DBAttributeClass()
        R.table                                 = self.tablealias
        R.attr                                  = self.attributename
        R.type                                  = self.dbType
        R.nosearch                              = self.nosearch
        R.rsindex                               = self.rsindex
        return R

    #-------------------------------------------------------------------------------------------------------
    def setDottedLocator( self, value ):
        raise NotImplementedError

    #-------------------------------------------------------------------------------------------------------
    def getDottedLocator( self ):
        R = [ self.tablealias ]
        if self.fktargetalias:
            R.append( self.fktargetalias )
        R.append( self.attributename )
        return '.'.join( R )
    #dottedLocator = property( **f() ); del f

    #-------------------------------------------------------------------------------------------------------
    def f( this, dBAttributeObject, id = 'UNKNOWN' ):
        R = this(
            id                                  =  id,
            tablealias                          =  dBAttributeObject.table,
            attributename                       =  dBAttributeObject.attr,
            dbType                              =  dBAttributeObject.type,
            nosearch                            =  dBAttributeObject.nosearch,
            rsindex                             =  dBAttributeObject.rsindex,
            )
        return R
    fromDBAttributeObject = classmethod( f ); del f

    #-------------------------------------------------------------------------------------------------------
    def refersToSameAttributeAsDBAttributeObject( self, dBAttributeObject ):
        R = ( True
            and self.tablealias                 == dBAttributeObject.table
            and self.attributename              == dBAttributeObject.attr
            #and self.dbType                     == dBAttributeObject.type
            #and self.nosearch                   == dBAttributeObject.nosearch
            #and self.rsindex                    == dBAttributeObject.rsindex
            )
        return R

    #-------------------------------------------------------------------------------------------------------
    def f( this, dBAttributeObject0, dBAttributeObject1 ):
        R = ( True
            and dBAttributeObject0.table        == dBAttributeObject1.table
            and dBAttributeObject0.attr         == dBAttributeObject1.attr
            #and dBAttributeObject0.type         == dBAttributeObject1.type
            #and dBAttributeObject0.nosearch     == dBAttributeObject1.nosearch
            #and dBAttributeObject0.rsindex      == dBAttributeObject1.rsindex
            )
        return R
    dBAttributeObjectsReferToSameAttribute = classmethod( f ); del f

############################################################################################################
#
#
#
#===========================================================================================================
class DsatDescriptions( list ):
    """A derivative of ``list`` for instances of ``DsatDescription`` that respects their ``~.sortentry``
    property when being sorted. The aim is to have a list that uniformly sorts data source attributes
    according to their table aliases, foreign key table alias target, and the name of the attribute itself
    for improved display."""

    def __init__( self, *P, **Q ):
        super( DsatDescriptions, self ).__init__( *P, **Q )
        self.idCache = {}

    #-------------------------------------------------------------------------------------------------------
    def sort( self ):
        R = [ ( description.sortentry, description ) for description in self ]
        R.sort()
        del self[:]
        for description.sortentry, description in R:
            self.append( description )

    #-------------------------------------------------------------------------------------------------------
    def clearIdCache( self ):
        self.idCache.clear()

    #-------------------------------------------------------------------------------------------------------
    def _buildIdCacheIfNecessary( self ):
        if len( self ) != len( self.idCache ):
            self.clearIdCache()
            for description in self:
                self.idCache[ description.id ] = description

    #-------------------------------------------------------------------------------------------------------
    def iterTablealiases( self ):
        for i, tablealias in self.iterEnumeratedTablealiases():
            yield tablealias

    #-------------------------------------------------------------------------------------------------------
    def iterEnumeratedTablealiases( self ):
        _lastTablealias = None
        i = -1
        for description in self:
            tablealias = description.tablealias
            if _lastTablealias == tablealias: continue
            _lastTablealias = tablealias
            i += 1
            yield i, tablealias

    #-------------------------------------------------------------------------------------------------------
    def iterDsatDescriptionsForTablealias( self, tablealias ):
        for description in self:
            thisTablealias = description.tablealias
            if tablealias != thisTablealias: continue
            yield description

    #-------------------------------------------------------------------------------------------------------
    def getDsatDescriptionForIdOrNone( self, id ):
        self._buildIdCacheIfNecessary()
        return self.idCache.get( id, None )

############################################################################################################
#
#
#
#===========================================================================================================
if __name__ == '__main__':

    sep  = '#' * 108
    sep1 = '=' * 108
    sep2 = '-' * 108
    import sys; sys.setrecursionlimit(50)


    descriptions = DsatDescriptions()

    descriptions.append(
        DsatDescription( id = 'bar', tablealias = 'tablealias300', attributename = 'attrn1' ) )
    descriptions.append(
        DsatDescription( id = 'bar', tablealias = 'tablealias300', attributename = 'attrn1', fktargetalias = 'tablealias712' ) )
    descriptions.append(
        DsatDescription( id = 'bar', tablealias = 'tablealias300', attributename = 'attrn1', fktargetalias = 'tablealias710' ) )
    descriptions.append(
        DsatDescription( id = 'foo', tablealias = 'tablealias123', attributename = 'attrn0' ) )
    descriptions.sort()
    for description in descriptions:
        print description

    print
    print sep
    for tablealias in descriptions.iterTablealiases():
        print sep1
        print tablealias
        for description in descriptions.iterDsatDescriptionsForTablealias( tablealias ):
            print description


