#
# -*- coding: utf-8 -*-

"""

Services and standards for Unicode handling.

Problem
============================================================================================================

Many things want to go wrong when working with both ordinary strings (octets) and Unicode objects employing
a variety of encodings and designing objects that produce textual representations. Many times components
that used to work fine balk as soon as they are fed a German umlaut or any character outside the scope of
US-ACII. Typically, you get long and involved exception stack traces with a bottom line like ::

    UnicodeEncodeError: 'ascii' codec can't encode character '\ue4' in position 10: ordinal not in range(128)

or similar. This snippet attempts to propose some fixes to avoid the dreaded encoding-related exceptions.


Solution
============================================================================================================


Source File Encoding
------------------------------------------------------------------------------------------------------------

First of all, it pays to be clear about the encoding of one's own source files. It is absolutely mandatory
to produce correctly encoded sourcefiles when working with Unicode and encodings, or else all will not go
well. It is probably a good idea to settle on a single encoding to be used throughout absolutely all
sources, and probably the best choices here are US-ASCII for its simplicity and universal acceptance, and
UTF-8, for its relative clarity, wide acceptance, and, above all, its universality when it comes to handling
characters outside the scope of US-ASCII. Best is, then, to include a magic comment on the second line of
the source file, like this::

    #!/path/to/python
    # -*- coding: utf-8 -*-

Users may feel at an advantage when prepending files with a Unicode UTF-8 Byte Order Mark (BOM), which will
allow compliant applications such as Python or some text editors to guess the encoding with great certainty.
I have, however, found that the Apache web server even in version 2 does *not* accept BOMs for CGI files,
since it assumes that the hash of the Shebang line is found in the first *byte* of the file, not in the
first regular *character*. This experience led me to switching off BOMs in my text editor.


Project-Wide Default Encoding
------------------------------------------------------------------------------------------------------------

Having decided about a standard for source encoding, it then becomes necessary to do the same for the
project (the module, the package). It is always possible or, depending on circumstances, likely that quaint
accented characters have to find their way into and out of the data objects the application, and when text
travels over the network or stored in a file, it becomes inevitable to express yourself in terms of octets.
It is then a bad idea to start all over in all the many places where such a thing may happen and re-write
decoding and encoding and testing routines. Worst of all, there may be quite a few places (for example,
wherever objects deliver a textual representation of their component data items) where a decision about a
reasonable encoding has to be made. It then becomes inevitable ...

make module with configuration

keep ``standardEncoding`` or similar there

prefer to make it same as source file encoding

if more than one encoding needed, build hierarchical configuration container


Standardize Standard Output
------------------------------------------------------------------------------------------------------------

cannot change system's default encoding since sys.setdefaultencoding is made unavailable; it would be a bad
idea anyhow

use redirection of sys.stdout -- since both strings and unicode objects are present there (and nothing
else), good point to ensure that at least all unguarded print statements that work with unicode objects work
fine


Design Serialization Behavior of Classes in a Unicode-Aware Manner
------------------------------------------------------------------------------------------------------------

used to be: often possible to avoid __str__ and __unicode__, only define __repr__ to display instance's
textual representation. now mandatory: always consciously define all three methods -- __str__, __unicode__,
and __repr__, -- of all classes that are meant to display more than <Xy instance at 1234> at the print
statement.


Make Yourself Heard: Publish Standards Globally
------------------------------------------------------------------------------------------------------------

Problem: settings made in one place not readily available in other places. need import, but then the other
side has to know you first, impossible. Other way round: place symbols in __builtin__.dict, then everyone
can read them who knows that public name, or manipulate global objects like sys.stdout. The problem then
arises of how to keep track of all these global names. This difficulty is addressed by the module
``configtool.utilities.globalNamebase``, which contains an object ``g`` that acts as a central module
registry. Specifically, ``IO`` needs a setting ``g.cfg.standardEncoding``.


Keep Clear Strings
------------------------------------------------------------------------------------------------------------

When defining texts with characters outside of US-ASCII, *always* use Unicode objects. Although Python does
accept things like ``s="äöü"`` with the source file encoding set to ``cp1252``, ``utf-8``, or ``latin-1``,
it is still vastly better not to put characters that go beyond the first 128 positions into octets.

Even after installing the standard output encoder, ``unicode( u'ä' )`` works, but ``unicode( 'ä' )`` still
miserably fails with a ``UnicodeDecodeError`` (if ``sys.setdefaultencoding()`` was not called or called with
the default setting, ``us-ascii``).


Global Namespace Reminder
============================================================================================================

This module relies on the presence of a global, quasi-builtin symbol ``g`` that acts as a project-wide
functionality proxy. See the module ``globalNamebase`` for details on how to construct and provide such a
symbol.

"""


############################################################################################################
#
#   STANDARD IMPORTS
#
#===========================================================================================================
import sys
import __builtin__


############################################################################################################
#
#   YELLOW PAGES
#
#===========================================================================================================
cfg                         = g.cfg

############################################################################################################
#
#
#
#===========================================================================================================
class FileOutputDiverterABC:

    def install( self ):
        self.previousStdout, sys.stdout = sys.stdout, self

    def remove( self ):
        sys.stdout = self.previousStdout

#===========================================================================================================
class StdOutEncoder( FileOutputDiverterABC ):

    def __init__( self ):
        self.file = sys.stdout
        self.previousStdout = None

    def __getattr__( self, attr ):
        return getattr( self.__dict__[ 'file' ], attr )

    def write( self, text ):
        if isinstance( text, unicode ):
            text = text.encode( cfg.defaultEncoding )
        self.file.write( text )

#===========================================================================================================
def installStdOutEncoder():
    StdOutEncoder().install()

#===========================================================================================================
def removeStdOutEncoder():
    if isinstance( sys.stdout, StdOutEncoder ):
        sys.stdout.remove()

#===========================================================================================================
def removeAllStdOutEncoders():
    while isinstance( sys.stdout, StdOutEncoder ):
        sys.stdout.remove()


############################################################################################################
#
#
#
#===========================================================================================================
def toUnicode( obj ):
    if   isinstance( obj, unicode ):    return obj
    elif isinstance( obj, str     ):    return obj.decode( cfg.defaultEncoding )
    else:
        try:                            return unicode( obj )
        except UnicodeDecodeError:      return unicode( str( obj ) )

#===========================================================================================================
def toOctets( obj ):
    if   isinstance( obj, str     ):    return obj
    elif isinstance( obj, unicode ):    return obj.encode( cfg.defaultEncoding )
    else:
        try:                            return str( obj )
        except UnicodeDecodeError:      return str( unicode( obj ) )

############################################################################################################
#
#
#
#===========================================================================================================
class Unicode:
    """Suitable ``~.__unicode__()`` methods must avoid redirection to ``~.__repr__()``, ``~.__str__()``, and
    ``toUnicode()``, since these routines (may) call ``~.__unicode__()`` themselves.

    It is no problem if ``~.__unicode__()`` atomizes its constituents in a suitable way and calls
    ``unicode()`` or ``toUnicode()`` for each element. ::

        def __unicode__( self ):
            return u'hello world'

    """
    class __str__:
        def __str__( self ):
            return self.__unicode__().encode( cfg.defaultEncoding )

    class __repr__:
        def __repr__( self ):
            return repr( self.__unicode__() )

    class __repr__2:
        def __repr__( self ):
            R = repr( self.__unicode__() )
            if R.startswith( 'u' ): R = R[1:]
            return R[1:-1]
#           def __repr__( self ):
#               return str( repr( self.__unicode__() ) )[1:-1] # this prevents ticks from showing up

    class StringPresenter( __str__, __repr__ ):
        pass

    class StringPresenter2( __str__, __repr__2 ):
        pass


############################################################################################################
#
#
#
#===========================================================================================================
if __name__ == '__main__':

    # Yellow Pages:
    IO = g.IO

    class X( IO.Unicode.StringPresenter ):

        def __unicode__( self ):
            return u'this is X ä'

    x = X()

    IO.installStdOutEncoder()

    print x
    print `x`
    print (1,2,3)
    print 'ä'
    print u'ä'

