#
# -*- coding: utf-8 -*-


"""


-- normalize inputs: turn termnumbers into strings, convert back on return

-- step through hints, fetching from each what is possible / missing


*analyze* an ontonym:
    turn the hints passed as arguments into a tuple ``(graphhint,locator)``, where
    ``graphhint`is either a graphname or a gauge (i.e. a graphname surrounded by colons) and locator is either a
    keyword, a path, or a termnr that may be either numerical or a string.

*resolve* an ontonym:
    deliver the object that the ontonym refers to, i.e. a node object or similar.


Ontonyms are identified by a single hint or a sequence of hints.

A hint may potentially be of the following types: an integer, a string, or an object that has been
registered for use as a gauge or a node reference. Assume ``graph`` to stand in for an object whose
attribute ``graph.gauge`` yields a gauge, and ``node`` to stand in for an object whose attribute
``node.graph`` yields an object of the same class as ``graph``, whose attribute ``node.locator`` yields a
locator, and whose attribute ``node.termnr`` yields a term nr (a numerical locator).

Sequences of hints are identified by calling a registered function, call it ``isHintSequence()`` that, when
presented with a legal input, returns True for sequences and False for single arguments.

Partially legal ontonyms are sequences of hints that, when considered from the first up to n-th element, at
some point complete the information needed to identify a term in an ontology.

Resolving an ontonym can then be accomplished in the following way: we first try to determine whether the
argument is a single hint or a sequence of hints. If it is a single hints, then we must get the information
on the ontology and the location of the termn in that ontology all from that argument.

Resolving a single hint works like this: It will only be possible to resolve a single argument if the
argument is either a node object( in which case we are finished, and the output is identical to the input),
or if it is a string spells out the location of the node searched for. Such a string, having to be complete,
will always have to start with a gauge, and, hence, only strings starting with ``:`` (a colon) are
acceptable in the first place. Following the initial colon must come a valid XML element type name, except
that the name of an ontology itself may not contain any colon. That part must be finished by a colon, and be
followed by at least one more character that is not a colon, but either a slash, or a latin-1-digit (that
is, match ``/[0-9]/``) or a character that is a legal first character of XML elemnt type names.



#   ::
#       forms of an ontonym:
#       '/locator'
#       ':ontology:/locator'
#       '42'
#       42
#       ':ontology:42' # ':ontology:42/locator' ???
#       ( 'ontology', 42 )
#       ( 'ontology', '42' )
#       ( 'ontology', '/locator' )
#       ( 'ontology', 42, context )
#       ( 'ontology', '42', context )
#       ( 'ontology', '/locator', context )
#       ( 42, context )
#       ( '42', context )
#       ( '/locator', context )
#
#       roughly, ontonym = ( [ gauge ], locator [, context ] )
#
#       context may be a node, a graph, or a string from which a graph is derivable (we do not use
#       node contexts, and node is reduced to node.graph here)

"""




############################################################################################################
#
#   YELLOW PAGES
#
#===========================================================================================================
#   from __future__ import division
#-----------------------------------------------------------------------------------------------------------
u                                       = g.configtool.utilities
#-----------------------------------------------------------------------------------------------------------
_Graph                                  = u.EzGraph.Graph
_Node                                   = u.EzGraph.Node
#-----------------------------------------------------------------------------------------------------------
COLON                                   = ':'
SLASH                                   = '/'
DOUBLESLASH                             = '//'

############################################################################################################
#
#   EXCEPTIONS
#
#===========================================================================================================
class OntonymAnalyzerError( Exception ): pass
class EmptyHintsequenceError( OntonymAnalyzerError, ValueError ): pass
class IllegalLocatorError( OntonymAnalyzerError, ValueError ): pass


############################################################################################################
#
#   ONTONYM ANALYZER
#
#===========================================================================================================
class OntonymAnalyzer( object ):

    #-------------------------------------------------------------------------------------------------------
    graphClass              = _Graph
    nodeClass               = _Node
    termnrsAsIntegers       = True

    #-------------------------------------------------------------------------------------------------------
    def __init__( self, termnrsAsIntegers = None ):
        if termnrsAsIntegers is not None:
            self.termnrsAsIntegers = termnrsAsIntegers

    #-------------------------------------------------------------------------------------------------------
    def analyze( self, hint, termnrsAsIntegers = None ):
        if self.isHintSequence( hint ):
            graphname, locator = self.analyzeMultipleHints( hint )
        else:
            graphname, locator = self.analyzeSingleHint( hint )
        return self._normalize( graphname, locator, termnrsAsIntegers )

    #-------------------------------------------------------------------------------------------------------
    def analyzeSingleHint( self, hint ):
        #---------------------------------------------------------------------------------------------------
        if self.isHintingObject( hint ):
            assert self.isNode( hint )
            return self.analyzeNode( hint )
        #---------------------------------------------------------------------------------------------------
        if self.isText( hint ):
            assert hint.startswith( COLON )
            secondColonPosition = hint.find( COLON, 1 )
            assert secondColonPosition > -1
            graphname = hint[ 1 : secondColonPosition ]
            assert self.isLegalGraphname( graphname )
            locator = hint[ secondColonPosition + 1 : ]
            #   skipNumericalTest = False
            #   if self.termnrsAsIntegers:
            #       if locator.isdigit(): return graphname, self.castTermnr( locator )
            #       skipNumericalTest = True
            #   self._assertLooksLikeLegalLocator( locator, skipNumericalTest )
            self._assertLooksLikeLegalLocator( locator )
            return graphname, locator
        #---------------------------------------------------------------------------------------------------
        raise NotImplementedError()

    #-------------------------------------------------------------------------------------------------------
    def analyzeNode( self, node ):
        graphname, locator = self.getGraphnameForNode( node ), self.getLocatorForNode( node )
        assert graphname is not None and locator is not None
        return graphname, locator

    #-------------------------------------------------------------------------------------------------------
    def analyzeText( self, text ):
        if text.startswith( COLON ):
            if text.endswith( COLON ):
                return text[ 1 : -1 ], None
            secondColonPosition = text.find( COLON, 1 )
            assert secondColonPosition > -1
            graphname   = text[ 1 : secondColonPosition ]
            locator     = text[ secondColonPosition + 1 : ]
            assert graphname
            assert locator
            return graphname, locator
        else:
            assert COLON not in text
            return None, text

    #-------------------------------------------------------------------------------------------------------
    def analyzeMultipleHints( self, hints ):
        #---------------------------------------------------------------------------------------------------
        if len( hints ) == 0:
            raise EmptyHintsequenceError( ''
                +   "Need at least one hint in hint sequence: %r." % ( hints, )
                )
        #---------------------------------------------------------------------------------------------------
        if len( hints ) == 1: return self.analyzeSingleHint( hints[ 0 ] )
        #---------------------------------------------------------------------------------------------------
        if len( hints ) > 2:
            raise HintsequenceTooLongError( ''
                +   "Unable to accept hint sequences with more than two elements: %r." % ( hints, )
                )
        #---------------------------------------------------------------------------------------------------
        hint0, hint1 = hints
        graphname0, locator0 = None, None
        graphname1, locator1 = None, None
        #---------------------------------------------------------------------------------------------------
        if self.isHintingObject( hint0 ):
            if self.isNode( hint0 ):
                return self.analyzeNode( hint0 )
            graphname0 = self.getGraphnameForGraph( hint0 )
        else:
            #   probably wrong here -- hint0 could be numeric
            graphname0, locator0 = self.analyzeText( hint0 )
        #---------------------------------------------------------------------------------------------------
        if graphname0 and locator0: return graphname0, locator0
        assert not ( graphname0 is None and locator0 is None )
        #---------------------------------------------------------------------------------------------------
        if self.isHintingObject( hint1 ):
            if self.isNode( hint1 ):
                if not graphname0: return self.getGraphnameForNode( hint1 ), locator0
                return graphname0, self.getLocatorForNode( hint1 )
            assert locator0 is not None
            return self.getGraphnameForGraph( hint1 ), locator0
        else:
            if self.isNumber( hint1 ):
                assert graphname0 is not None
                return graphname0, hint1
            graphname1, locator1 = self.analyzeText( hint1 )
            if not graphname0:
                assert locator0 is not None
                assert graphname1 is not None
                return graphname1, locator0
            assert graphname0 is not None
            assert locator1 is not None
            return graphname0, locator1
        raise NotImplementedError()

    #-------------------------------------------------------------------------------------------------------
    #   STRING ANALYSIS
    #-------------------------------------------------------------------------------------------------------
    def _assertLooksLikeLegalLocator( self, text ):
        if text.isdigit(): return
        if text.startswith( SLASH ) and len( text ) < 2:
            raise IllegalLocatorError( ''
                +   "A locator cannot consist of a single slash: %r" % ( text, )
                )
        if self._containsWhitespace( text ):
            raise IllegalLocatorError( ''
                +   "A locator cannot contain whitespace: %r" % ( text, )
                )
        if DOUBLESLASH in text:
            raise IllegalLocatorError( ''
                +   "A locator cannot contain empty steps: %r" % ( text, )
                )

    #-------------------------------------------------------------------------------------------------------
    def _containsWhitespace( self, text ):
        return len( text.split() ) > 1

#       #-------------------------------------------------------------------------------------------------------
#       def _containsOnlyLegalLocatorCharacters( self, locator ):
#           raise NotImplementedError()

#       #-------------------------------------------------------------------------------------------------------
#       def _tryToSplitGaugeLocator( self, text ):
#           if not self._looksLikeStartswithGauge( text ): return None, None
#           self.analyzeSingleHint( a )
#           try:
#               return self._assertLooksLikeLegalLocator( text[

    #-------------------------------------------------------------------------------------------------------
    def _looksLikeStartswithGauge( self, text ):
        if not len( text ) > 2: return False
        if not text.startswith( COLON ): return False
        if not text.count( COLON ) == 2: return False

    #-------------------------------------------------------------------------------------------------------
    def _looksLikeGauge( self, text ):
        if not self._looksLikeStartswithGauge( text ): return False
        return text.endswith( COLON )

    #-------------------------------------------------------------------------------------------------------
    def isLegalGraphname( self, graphname ):
        return True

    #-------------------------------------------------------------------------------------------------------
    #   ATTRIBUTE RETRIEVAL
    #-------------------------------------------------------------------------------------------------------
    def getGaugeForObject( self, a ):
        if isinstance( a, self.graphClass ):    return self.getGaugeForGraph( a )
        if isinstance( a, self.nodeClass ):     return self.getGaugeForNode( a )

    #-------------------------------------------------------------------------------------------------------
    def getGraphnameForObject( self, a ):
        if isinstance( a, self.graphClass ):    return self.getGraphnameForGraph( a )
        if isinstance( a, self.nodeClass ):     return self.getGraphnameForNode( a )

    #-------------------------------------------------------------------------------------------------------
    def getLocatorForObject( self, a ):
        if isinstance( a, self.nodeClass ):     return self.getLocatorForNode( a )

    #-------------------------------------------------------------------------------------------------------
    def getTermnrForObject( self, a ):
        if isinstance( a, self.nodeClass ):     return a.termnr

    #-------------------------------------------------------------------------------------------------------
    def getGraphnameForGraph( self, graph ):    return graph.name
    def getGaugeForGraph( self, graph ):        return ':%s:' % ( self.getGraphnameForGraph( graph ), )
    def getLocatorForNode( self, node ):        return node.locator
    def getGraphnameForNode( self, node ):      return self.getGraphnameForGraph( node.graph )
    def getGaugeForNode( self, node ):          return self.getGaugeForGraph( node.graph )

    #-------------------------------------------------------------------------------------------------------
    #   NORMALIZATION
    #-------------------------------------------------------------------------------------------------------
    def _normalize( self, graphname, locator, termnrsAsIntegers = None ):
        assert graphname is not None
        assert locator   is not None
        if termnrsAsIntegers is None: termnrsAsIntegers = self.termnrsAsIntegers
        if self.isText( locator ):
            if termnrsAsIntegers and self.isNumericaltext( locator ):
                locator = self.castTextTermnrToNumber( locator )
        elif not termnrsAsIntegers:
            locator = self.castNumericalTermnrToText( locator )
        return graphname, locator

#           if termnrsAsIntegers and self.isText( locator ) and locator.isdigit():
#               return graphname, self.castTextTermnrToNumber( locator )
#           if not termnrsAsIntegers and locator.isdigit():
#               return graphname, self.castNumericalTermnrToText( locator )
#           return graphname, locator

    #-------------------------------------------------------------------------------------------------------
    def castTextTermnrToNumber( self, text ):
        return int( text )

    #-------------------------------------------------------------------------------------------------------
    def castNumericalTermnrToText( self, number ):
        return '%d' % ( number, )

    #-------------------------------------------------------------------------------------------------------
    #   TYPE ANALYSIS
    #-------------------------------------------------------------------------------------------------------
    def isHintSequence( self, a ):
        return isinstance( a, ( list, tuple, ) )

    #-------------------------------------------------------------------------------------------------------
    def isText( self, a ):
        return isinstance( a, basestring )

    #-------------------------------------------------------------------------------------------------------
    def isNumber( self, a ):
        return isinstance( a, ( int, long ) )

    #-------------------------------------------------------------------------------------------------------
    def isNumericaltext( self, text ):
        return text.isdigit()

    #-------------------------------------------------------------------------------------------------------
    def isHintingObject( self, a ):
        return isinstance( a, ( self.graphClass, self.nodeClass, ) )

    #-------------------------------------------------------------------------------------------------------
    def isGraph( self, a ):
        return isinstance( a, self.graphClass )

    #-------------------------------------------------------------------------------------------------------
    def isNode( self, a ):
        return isinstance( a, self.nodeClass )


############################################################################################################
#
#
#
#===========================================================================================================
if __name__ == '__main__':

    import sys; sys.setrecursionlimit( 50 )
    sep0 = '#' * 108
    sep1 = '=' * 108
    sep2 = '-' * 108

    graph0 = _Graph( 'on' )
    nodeLo = graph0.newNodeForLocator( '/lo' )
    nodeLoCa = graph0.newNodeForLocator( '/lo/ca' )
    nodeLoCaTor = graph0.newNodeForLocator( '/lo/ca/tor' )
    oa = OntonymAnalyzer()
    #-------------------------------------------------------------------------------------------------------
    probeResultPairs = [
        #---------------------------------------------------------------------------------------------------
        ( nodeLoCaTor,                      ( 'on', '/lo/ca/tor', )                     ),
        ( ':on:keyword',                    ( 'on', 'keyword', )                        ),
        ( ':on:/lo/ca/tor',                 ( 'on', '/lo/ca/tor', )                     ),
        ( ':on:42',                         ( 'on', 42, )                               ),
        #---------------------------------------------------------------------------------------------------
        ( ( ':on:keyword', ),               ( 'on', 'keyword', )                        ),
        ( ( ':on:/lo/ca/tor', ),            ( 'on', '/lo/ca/tor', )                     ),
        ( ( ':on:42', ),                    ( 'on', 42, )                               ),
        #---------------------------------------------------------------------------------------------------
        ( ( ':on:', 'keyword', ),           ( 'on', 'keyword', )                        ),
        ( ( ':on:', '/lo/ca/tor', ),        ( 'on', '/lo/ca/tor', )                     ),
        ( ( ':on:', '42', ),                ( 'on', 42, )                               ),
        ( ( ':on:', nodeLoCaTor, ),         ( 'on', '/lo/ca/tor', )                     ),
        ( ( graph0, nodeLoCaTor, ),         ( 'on', '/lo/ca/tor', )                     ),
        ( ( graph0, '42', ),                ( 'on', 42, )                               ),
        ( ( graph0, 42, ),                  ( 'on', 42, )                               ),
        ( ( graph0, '/lo/ca/tor', ),        ( 'on', '/lo/ca/tor', )                     ),
        ]
        #---------------------------------------------------------------------------------------------------

    for probe, expectedResult in probeResultPairs:
        print '%30r == %r ?' % ( probe, expectedResult, ),
        actualResult = oa.analyze( probe )
        if not actualResult == expectedResult:
            assert False, "\nResult violates expectation: %r" % ( actualResult, )
        print 'ok'


    #-------------------------------------------------------------------------------------------------------
    probeResultPairs = [
        ( ':on:42',                         ( 'on', '42', )                               ),
        ( ( ':on:42', ),                    ( 'on', '42', )                               ),
        ( ( ':on:', '42', ),                ( 'on', '42', )                               ),
        ( ( graph0, '42', ),                ( 'on', '42', )                               ),
        ( ( graph0, 42, ),                  ( 'on', '42', )                               ),
        ]
        #---------------------------------------------------------------------------------------------------

    for probe, expectedResult in probeResultPairs:
        print '%30r == %r ?' % ( probe, expectedResult, ),
        actualResult = oa.analyze( probe, termnrsAsIntegers = False )
        if not actualResult == expectedResult:
            assert False, "\nResult violates expectation: %r" % ( actualResult, )
        print 'ok'

