#
# -*- coding: utf-8 -*-


"""Commonly used idioms dealing with files, filenames, paths, and directories."""


import sys
import inspect
import os
import os.path as ospath
import shutil

def filename( obj ):
    """Return filename where ``obj`` is defined; this will commonly be the ``*.py`` or the ``*.pyc`` file
    contain the (byte-compiled) source of the file where ``obj`` is defined. Typically, one will construct a
    dummy object on the fly for the argument and write, e.g., ``filename(lambda:None)``."""
    return inspect.getfile( obj )

def here( obj ):
    """Return the name of the directory of the filename returned by ``filename(obj)``."""
    return ospath.dirname( filename( obj ) )
dirname = here

def includeThisDirectoryInPath( obj ):
    """Prepend the name of the directory returned by ``here(obj)`` to ``sys.path``, so that subsequent
    ``import`` statements act in a localized fashion. This is useful in situations where a script is not
    called in the usual way (such as when using ``execfile()`` but it must be ensured that local imports
    work."""
    sys.path.insert( 0, here( obj ) )


#   def readb( *P ):
#       """Like ``read()``, q.v., but returning binary contents of a file."""
#       fso = file( join( *P ), 'rb' )
#       R = fso.read()
#       fso.close()
#       return R

def read( *P ):
    """Assuming that ``P`` contains filelocator components, assemble pieces using ``ospath.join`` and
    return contents of the file identified by the resulting locator."""
    return _read( 'r', 'read', *P )

def readlines( *P ):
    """Assuming that ``P`` contains filelocator components, assemble pieces using ``ospath.join`` and
    return contents of the file identified by the resulting locator."""
    return _read( 'r', 'readlines', *P )

def readplainlines( *P ):
    """Assuming that ``P`` contains filelocator components, assemble pieces using ``ospath.join`` and
    return contents of the file identified by the resulting locator."""
    return [ line[ : -1 ] for line in readlines( *P ) ]

def readBinary( *P ):
    """Assuming that ``P`` contains filelocator components, assemble pieces using ``ospath.join`` and
    return contents of the file identified by the resulting locator."""
    return _read( 'rb', 'read', *P )

def readlinesBinary( *P ):
    """Assuming that ``P`` contains filelocator components, assemble pieces using ``ospath.join`` and
    return contents of the file identified by the resulting locator."""
    return _read( 'rb', 'readlines', *P )

def _read( mode, readMethodname, *P ):
    """Assuming that ``P`` contains filelocator components, assemble pieces using ``ospath.join`` and
    return contents of the file identified by the resulting locator."""
    fso = file( join( *P ), mode )
    R = getattr( fso, readMethodname )()
    fso.close()
    return R


def join( *P ):
    return ospath.join( *P )

def weld( *P ):
    if P:
        p = P[ 0 ]
        if not isinstance( p, basestring ):
            P = list( P )
            P[ 0 ] = here( p )
    return ospath.join( *P )

#===========================================================================================================
filter_ = filter
#-----------------------------------------------------------------------------------------------------------
def iterateNameLocatorPairs( directorylocator, filter = None ):
    #-------------------------------------------------------------------------------------------------------
    if filter is None:
        filter = lambda locator: True
    #-------------------------------------------------------------------------------------------------------
    for name, locator in [
        ( name, ospath.join( directorylocator, name ) )
            for name in os.listdir( directorylocator ) ]:
        if os.sep != '/':
            locator = locator.replace( os.sep, '/' )
        if filter( locator ):
            yield name, locator

#===========================================================================================================
def iterateNames( directorylocator, filter = None ):
    for nlp in iterateNameLocatorPairs( directorylocator, filter ):
        yield nlp[ 0 ]

#===========================================================================================================
def iterateLocators( directorylocator, filter = None ):
    for nlp in iterateNameLocatorPairs( directorylocator, filter ):
        yield nlp[ 1 ]

#===========================================================================================================
def iterateNameLocatorPairsOfDirectories( directorylocator, filter = None ):
    #-------------------------------------------------------------------------------------------------------
    if filter is None:
        filter = lambda locator: True
    #-------------------------------------------------------------------------------------------------------
    filter2 = \
        lambda locator: \
            ospath.isdir( locator ) and filter( locator )
    #-------------------------------------------------------------------------------------------------------
    for name, locator in iterateNameLocatorPairs( directorylocator, filter2 ):
        yield name, locator


#===========================================================================================================
def safewrite( *P, **Q ):
    """source-safe file handling procedure:

    policy: keep backups 'on both ends'.

    steps:

        1)  before writing, create a backup of file.
            (this could also happen in several stages):

            foo.xml         => foo.xml.backup

        2)  when writing, write to a 'forward' of file:

            data            => foo.xml.forward

        3)  when all went ok, try to copy forward file
            over actual file:

            foo.xml.forward => foo.xml

    if anything goes wrong anywhere, backup file is kept
    with previous content.

    if write process bails out, the actual file is left untouched.

    if actual file is blocked because of concurrent access, new
    file contents are preserved in forward file."""
    raise NotImplementedError

#===========================================================================================================
def getForwardFilename( filename ):
    """Return ``filename`` plus the extension ``.forward``, which may be used to perform a tentative
    write operation."""
    return '%s.forward' % ( filename, )

#===========================================================================================================
def getBackupFilename( filename ):
    """Return ``filename`` plus the extension ``.backup``."""
    return '%s.backup' % ( filename, )

#===========================================================================================================
def createBackup( filename ):
    """Create a backup of a file by the name of ``filename`` in the same directory, adding the extension
    ``.backup``, keeping modification dates."""
    shutil.copy2( filename, getBackupFilename( filename ) )

#===========================================================================================================
def applyForward( filename, removeForward = False ):
    """Copy forward file (by the name of ``%(filename)s.forward``) over file ``filename``, *without*
    producing a backup. In order for this to work correctly, the forward file must have been closed
    by the time this function is called."""
    forwardfilename = getForwardFilename( filename )
    if removeForward:
        if ospath.isfile( filename ):
            os.remove( filename )
        os.rename( forwardfilename, filename )
    else:
        shutil.copy2( forwardfilename, filename )

#===========================================================================================================
def applyForwardWithBackup( filename ):
    """Copy forward file (by the name of ``%(filename)s.forward``) over file ``filename`` after saving
    original file to ```%(filename)s.backup``."""
    createBackup( filename )
    applyForward( filename )


"""

Script that establishes a lock::

    #   file: filelocking.py

    fsi = g.configtool.utilities.filesystemIdioms
    locator = 'c:/temp/lockedfile.txt'

    fsi.establishLock( locator )
    try:
        fsoW = file( locator, 'wb' )
        #fsoW.write( 'hello world' )
        fsoW.flush()
        for i in range( long( 5 * 10e5 ) ):
            fsoW.write( 'o' )
        fsoW.flush()
        fsoW.close()
    finally:
        fsi.releaseLock( locator )


Script that waits for the lock to be released::

    #   file: filelockingB.py

    fsi = g.configtool.utilities.filesystemIdioms
    locator = 'c:/temp/lockedfile.txt'

    def fLocked(): print 'locked...'
    def fTimeout(): print 'timeout'

    fsi.waitForLockReleased( locator, timeout = 25, lockedCallback = fLocked, timeoutCallback = fTimeout )
    print 'reading file...'
    fsoR = file( locator, 'rb' )
    bytes = fsoR.read()
    print len( bytes )
    #print `bytes`

"""

class FilelockerException( IOError ): pass
class FilelockerTimeoutError( IOError ): pass

#===========================================================================================================
def getLockfileLocator( locator ):
    return '%s.locked' % locator

#===========================================================================================================
def establishLock( locator, magic = 'lock' ):
    lockfilelocator = getLockfileLocator( locator )
    _establishLock( lockfilelocator, magic )

#===========================================================================================================
def _establishLock( lockfilelocator, magic = 'lock' ):
    _waitForLockReleased( lockfilelocator )
    lockfile = file( lockfilelocator, 'wb' )
    lockfile.write( magic )
    lockfile.flush()
    lockfile.close()
    del lockfile

#===========================================================================================================
def releaseLock( locator, magic = 'lock' ):
    lockfilelocator = getLockfileLocator( locator )
    _releaseLock( lockfilelocator )

#===========================================================================================================
def _releaseLock( lockfilelocator, magic = 'lock' ):
    from time import sleep
    if not ospath.isfile( lockfilelocator ): return
    try:
        trials = 0
        while True:
            trials += 1
            try:
                os.remove( lockfilelocator )
                if not ospath.isfile( lockfilelocator ): return
            except:
                if trials > 5: raise
                sleep( 0.07 )
    except Exception, e:
        import warnings
        warnings.warn( ''
            +   "Attempt to remove lockfile %s caused an exception to be raised: " % ( lockfilelocator, )
            +   "(%s) %s" % ( e.__class__, e, )
            )

#===========================================================================================================
def waitForLockReleased( locator, timeout = 5, lockedCallback = None, timeoutCallback = None ):
    lockfilelocator = getLockfileLocator( locator )
    _waitForLockReleased( lockfilelocator, timeout, lockedCallback, timeoutCallback )

#===========================================================================================================
def _waitForLockReleased( lockfilelocator, timeout = 5, lockedCallback = None, timeoutCallback = None ):
    from time import time, sleep
    t0 = time()
    while ospath.isfile( lockfilelocator ):
        if lockedCallback is not None:
            lockedCallback()
            lockedCallback = None
        sleep( 0.05 )
        if time() - t0 < timeout:
            continue
        else:
            if timeoutCallback is None:
                raise FilelockerTimeoutError(
                    "Timeout while waiting for lockfile %s to be removed" % ( lockfilelocator, ) )
            else:
                timeoutCallback()
        break

#===========================================================================================================
isfile = os.path.isfile
isdirectory  = os.path.isdir
exists  = os.path.exists

#   #===========================================================================================================
#   def filename( filelocator ):
#       CLASHES W/ PREVIOUS DEF
#       return ospath.basename( filelocator )



#   """
#
#   The present module, +ospath+, is a quick & dirty approach to unify in a way the manyfold options
#   present in Python to deal with directories and files. For the most part, these functions are
#   spread over the standard modules +os+ and +ospath+, plus a few built-ins. +ospath+ collects a
#   few of the more useful things from +os+ and +ospath+ and adds handy functions.
#
#   A more principled approach is certainly desirable. An imaginable hierarchy:
#
#   Fso (file system object):
#
#       <methods:>
#           exists
#           make
#           delete
#           read
#           write
#           chmod
#           rename
#
#       <derived classes:>
#           dir
#           file
#           link
#
#   It would be possible to establish an Fso object by giving one or more paths (or other Fsos); the
#   object could then be questioned whether it exists, what rights there are and so on.
#
#   """
#
#
#   #   #   from os-file-dir:
#   #
#   #       access(path, mode)
#   #       F_OK
#   #       R_OK
#   #       W_OK
#   #       X_OK
#   #       chdir(path)
#   #       getcwd()
#   #       chroot(path)
#   #       chmod(path, mode)
#   #       chown(path, uid, gid)
#   #       link(src, dst)
#   #       listdir(path)
#   #       lstat(path)
#   #       mkfifo(path[, mode])
#   #       mkdir(path[, mode])
#   #       makedirs(path[, mode])
#   #       pathconf(path, name)
#   #       pathconf_names
#   #       readlink(path)
#   #       remove(path)
#   #       removedirs(path)
#   #       rename(src, dst)
#   #       renames(old, new)
#   #       rmdir(path)
#   #       stat(path)
#   #       statvfs(path)
#   #       symlink(src, dst)
#   #       tempnam([dir[, prefix]])
#   #       tmpnam()
#   #       TMP_MAX
#   #       unlink(path)
#   #       utime(path, times)
#   #
#   #
#   #   from os-fd-ops:
#   #
#   #       close(fd)
#   #       dup(fd)
#   #       dup2(fd, fd2)
#   #       fpathconf(fd, name)
#   #       fstat(fd)
#   #       fstatvfs(fd)
#   #       ftruncate(fd, length)
#   #       isatty(fd)
#   #       lseek(fd, pos, how)
#   #       O_APPEND
#   #       O_BINARY
#   #       O_CREAT
#   #       O_DSYNC
#   #       O_EXCL
#   #       O_NDELAY
#   #       O_NOCTTY
#   #       O_NONBLOCK
#   #       O_RDONLY
#   #       O_RDWR
#   #       O_RSYNC
#   #       O_SYNC
#   #       O_TRUNC
#   #       O_WRONLY
#   #       open(file, flags[, mode])
#   #       openpty()
#   #       pipe()
#   #       read(fd, n)
#   #       tcgetpgrp(fd)
#   #       tcsetpgrp(fd, pg)
#   #       ttyname(fd)
#   #       write(fd, str)
#   #
#   #   from os-newstreams (File Object Creation):
#   #
#   #       fdopen(fd[, mode[, bufsize]])
#   #       popen(command[, mode[, bufsize]])
#   #       popen2(cmd[, mode[, bufsize]])
#   #       popen3(cmd[, mode[, bufsize]])
#   #       popen4(cmd[, mode[, bufsize]])
#   #       tmpfile()
#
#   #   from ospath:
#   #       abspath(path)
#   #       basename(path)
#   #       commonprefix(list)
#   #       dirname(path)
#   #       exists(path)
#   #       expanduser(path)
#   #       expandvars(path)
#   #       getatime(path)
#   #       getmtime(path)
#   #       getsize(path)
#   #       isabs(path)
#   #       isdir(path)
#   #       isfile(path)
#   #       islink(path)
#   #       ismount(path)
#   #       join(path1[, path2[, ...]])
#   #       normcase(path)
#   #       normpath(path)
#   #       realpath(path)
#   #       samefile(path1, path2)
#   #       sameopenfile(fp1, fp2)
#   #       samestat(stat1, stat2)
#   #       split(path)
#   #       splitdrive(path)
#   #       splitext(path)
#   #       walk(path, visit, arg)
#
#   import os
#   from os import chdir, getcwd, chmod, listdir, remove, mkdir, makedirs, rename, rmdir, \
#       removedirs, unlink, pardir, curdir, stat
#   from ospath import *
#
#   import shutil
#   from shutil import *
#
#   import inspect as _inspect
#
#   #   def normcase( path ):
#   #   regularize drive letter on windows
#
#   def absjoin( *paths ):
#       """Join +paths+ and turn them into an absolute path."""
#       return ospath.abspath( join( *paths ) )
#
#   def absdirfile( path ):
#       absfilename = ospath.abspath( path )
#       return truepath( absfilename ), absfilename
#
#   def truepath( path ):
#       return ospath.dirname( ospath.abspath( path ) )
#
#   def slashpath( path ):
#       """Do the inverse of +localseppath()+."""
#       return path.replace( os.sep, '/' )
#
#   def localseppath( path ):
#       """Return a copy of +path+ with slashes converted to separators as used on this machine."""
#       return path.replace( '/', os.sep )
#
#   def mydir( object ):
#       """Return the directory where +object+ was defined. Usage example::
#
#           here = ospath.mydir( lambda:0 )
#
#       returns the directory of the module where +mydir+ was called (even if that module was
#       imported by another module)."""
#       return truepath( myfilename( object ) )
#
#   def myfilename( object ):
#       return abspath( _inspect.getsourcefile( object ) )
#
#   def rmdir_recursive( dir, verbose = 0 ):
#       """Remove a directory, and all its contents if it is not already empty. -- Original code by
#       Ben Hutchings (ben.hutchings@roundpoint.com), posted on comp.lang.python on 2000-12-05
#       08:05:55 PST."""
#       for name in os.listdir( dir ):
#           full_name = ospath.join(dir, name)
#           # on Windows, if we don't have write permission we can't remove
#           # the file/directory either, so turn that on
#           if not os.access( full_name, os.W_OK ):
#               os.chmod( full_name, 0600 )
#           if ospath.isdir( full_name ):
#               rmdir_recursive( full_name, verbose )
#           else:
#               os.remove( full_name )
#               if verbose: print "removed %s" % slashpath( full_name )
#       os.rmdir( dir )
#       if verbose: print "removed %s" % slashpath( dir )
#
#   def copy( src, dst ):
#       """Use +shutil.copy2()+ or +shutil.copytree()+ to copy fiels and directories from +src+ to
#       +dst+. """
#       if ospath.isdir( src ):
#           shutil.copytree( src, dst )
#
#       elif ospath.isfile( src ):
#           shutil.copy2( src, dst )
#
#
#   def newest( *fsonames ):
#       """From a list of file system object names, return the one of the most recently modified
#       object. In case more than one file is electable, the order of arguments as passed to the
#       function determines precedence. If any of the names does not correspond to an existing
#       object or does not permit the application of +os.stat().st_mtime+, an +OSError+ is
#       raised."""
#       fsodates = map( lambda x: os.stat( x ).st_mtime, fsonames )
#       try:
#           return fsonames[ fsodates.index( max( fsodates ) ) ]
#       except ValueError:
#           pass
#
#   def existing( *fsonames ):
#       """From a list of names of putative file system object, return a list of those that actually
#       do correspond to existing objects."""
#       return filter( ospath.exists, fsonames )
#
#   def files( *fsonames ):
#       """From a list of names of putative file system object, return a list of names that
#       correspond to actually existing objects which are files."""
#       return filter( ospath.isfile, fsonames )
#
#   def newest_existing_filename( *fsonames ):
#       """From a list of names of putative file system object, return the name that corresponds to
#       the most recently modified file. This is implemented as a combination of +newest()+ and
#       +files()+. -- The name '+newest_existing_filename+' stresses the fact that the present function
#       does accept arbitrary names (in contradistinction to +newest()+)."""
#       return newest( *files( *fsonames ) )
#
#
#   def weld( object, *filenames ):
#       """Much like +join()+, except the very first argument should be an +object+ (such as a
#       module or an instance) whose place of birth is tractable by +mydir()+ (hence, it must be
#       responsive to +inspect.getsourcefile()+). This function comes in extremely practical
#       whenever double keeping of both import path and filesystem paths should be avoided -- just
#       place an +__init__.py+ file into your resources directory and place it into a directory on
#       the Python path (somewhere within your package, normally), et voila, data files may be
#       accessed as::
#
#           import myresources
#           myfile = file( ospath.weld( myresources, 'myfile.txt' ) )
#
#       """
#       return join( mydir( object ), join( *filenames ) )
#
#
#   if __name__ == '__main__':
#
#       mydir = mydir( lambda:0 )
#       print mydir
#       print absjoin( mydir, '..' )
#
#       try:
#           makedirs( "foo/bar/baz" )
#           mkdir( "foo/bar/blah" )
#           open( "foo/bar/foo.txt", 'w' )
#       except OSError:
#           pass
#       else:
#           rmdir_recursive( 'foo', verbose = 1 )
#
#
#   #       print newest( *existing( *files( 'c:/', 'xfoo.tbl', 'foo.tbl', 'recfldtable.py' ) ) )
#   #       print newest( *files( 'c:/', 'xfoo.tbl', 'foo.tbl', 'recfldtable.py' ) )
#   #       print newest( 'foo.tbl', 'recfldtable.py' )
#   #       print files( 'c:/', 'xfoo.tbl', 'foo.tbl', 'recfldtable.py' )
#   #
#   #       print newest_existing_filename( *os.listdir( '.' ) )
#   #
#   #
