#
# -*- coding: utf-8 -*-

"""

String interpolation services. This is a slight adaption of a module written by Ka-Ping Yee.

Updates
============================================================================================================

by Wolfgang Lipp, lipp@epost.de

2003-10-10:

    *   added crude testing code; this is all but finished.

    *   made to work with Unicode. Now possible to indicate a default encoding to be tried when other
        encodings fail. However, relying on this is still bad programming; instead, users should make sure
        they always use Unicode strings *whenever* there *may* be characters outside of US-ASCII. Also,
        internal string-to-Unicode-coercion within this module may not yet work correctly, which is to be
        checked.

    *   Deleted references to ``string`` module and type-checking of format string.

    *   Changed module and class name to ``kpyInterpol``.

    *   deleted convenience function for the time to enhance clarity of core concepts.




"""

#   Original Description
#   ============================================================================================================
#
#   Ka-Ping Yee's original module description (names changed)::
#
#       String interpolation for Python (by Ka-Ping Yee, 14 Feb 2000).
#
#       This module lets you quickly and conveniently interpolate values into
#       strings (in the flavour of Perl or Tcl, but with less extraneous
#       punctuation).  You get a bit more power than in the other languages,
#       because this module allows subscripting, slicing, function calls,
#       attribute lookup, or arbitrary expressions.  Variables and expressions
#       are evaluated in the namespace of the caller.
#
#       The inter() function returns the result of interpolating a string, and
#       printpl() prints out an interpolated string.  Here are some examples:
#
#           from kpyInterpol import printpl
#           printpl("Here is a $string.")
#           printpl("Here is a $module.member.")
#           printpl("Here is an $object.member.")
#           printpl("Here is a $functioncall(with, arguments).")
#           printpl("Here is an ${arbitrary + expression}.")
#           printpl("Here is an $array[3] member.")
#           printpl("Here is a $dictionary['member'].")
#
#       The filter() function filters a file object so that output through it
#       is interpolated.  This lets you produce the illusion that Python knows
#       how to do interpolation:
#
#           import kpyInterpol
#           sys.stdout = kpyInterpol.filter()
#           f = "fancy"
#           print "Isn't this $f?"
#           print "Standard output has been replaced with a $sys.stdout object."
#           sys.stdout = kpyInterpol.unfilter()
#           print "Okay, back $to $normal."
#
#       Under the hood, the kpyInterpol class represents a string that knows how to
#       interpolate values.  An instance of the class parses the string once
#       upon initialization; the evaluation and substitution can then be done
#       each time the instance is evaluated with str(instance).  For example:
#
#           from kpyInterpol import kpyInterpol
#           s = kpyInterpol("Here is $foo.")
#           foo = 5
#           print str(s)
#           foo = "bar"
#           print str(s)
#

import sys
import inspect
from tokenize import tokenprog

INTERESCAPE     = u"$"
OPENBRACE       = u"{"
CLOSEBRACE      = u"}"
OPENBRACKETS    = u"(["
CLOSEBRACKETS   = u")]"
DOT             = u"."
DEFAULT_ENCODING = sys.getdefaultencoding()

class InterpolationError(ValueError):
    def __init__(self, text, pos):
        self.text = text
        self.pos = pos
    def __str__(self):
        return u"Unfinished expression in %s at char %d." % (
            repr(self.text), self.pos)

def matchorfail(text, pos):
    match = tokenprog.match(text, pos)
    if match is None:
        raise InterpolationError(text, pos)
    return match, match.end()

class kpyInterpol:
    """Class representing a string with interpolation abilities.

    Upon creation, an instance works out what parts of the format
    string are literal and what parts need to be evaluated.  The
    evaluation and substitution happens in the namespace of the
    caller when str(instance) is called."""

    def __init__( self, format, recursive = False, defaultEncoding = DEFAULT_ENCODING ):
        """The single argument to this constructor is a format string.

        The format string is parsed according to the following rules:

        1.  A escPos sign and a name, possibly followed by any of:
              - an open-paren, and anything up to the matching paren
              - an open-bracket, and anything up to the matching bracket
              - a period and a name
            any number of times, is evaluated as a Python expression.

        2.  A escPos sign immediately followed by an open-brace, and
            anything up to the matching close-brace, is evaluated as
            a Python expression.

        3.  Outside of the expressions described in the above two rules,
            two escPos signs in a row give you one literal escPos sign."""

        self.defaultEncoding = defaultEncoding
        self.recursive = recursive
        if not isinstance( format, unicode ):
            format = format.decode( self.defaultEncoding )
        self.format = format

        namechars = u"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";
        chunks = []
        pos = 0

        while 1:
            escPos = format.find( INTERESCAPE, pos )
            if escPos < 0:
                break
            nextchar = format[ escPos + 1 ]

            if nextchar == OPENBRACE:
                chunks.append( ( 0, format[ pos : escPos ] ) )
                pos, level = escPos + 2, 1
                while level:
                    match, pos = matchorfail( format, pos )
                    tstart, tend = match.span( 3 )
                    token = format[ tstart : tend ]
                    if token == OPENBRACE:
                        level += 1
                    elif token == CLOSEBRACE:
                        level -= 1
                chunks.append( ( 1, format[ escPos + 2 : pos - 1 ] ) )

            elif nextchar in namechars:
                chunks.append( ( 0, format[ pos : escPos ] ) )
                match, pos = matchorfail( format, escPos + 1 )
                while pos < len( format ):
                    if format[ pos ] == DOT and \
                        pos+1 < len(format) and format[pos+1] in namechars:
                        match, pos = matchorfail(format, pos+1)
                    elif format[pos] in OPENBRACKETS:
                        pos, level = pos+1, 1
                        while level:
                            match, pos = matchorfail(format, pos)
                            tstart, tend = match.span(3)
                            token = format[tstart:tend]
                            if token[0] in OPENBRACKETS: level = level+1
                            elif token[0] in CLOSEBRACKETS: level = level-1
                    else:
                        break
                chunks.append((1, format[escPos+1:pos]))

            else:
                chunks.append((0, format[pos:escPos+1]))
                pos = escPos + 1 + ( nextchar == INTERESCAPE )

        if pos < len(format): chunks.append((0, format[pos:]))
        self.chunks = chunks

    def __str__( self ):
        return self.__unicode__().encode( self.defaultEncoding )

    def __repr__( self ):
        return repr( self.__unicode__() )

    def __unicode__( self ):
        """Evaluate and substitute the appropriate parts of the string."""
        callerLocals, callerGlobals = getCallerNamespaces( -1 )
        expansions = []
        for live, chunk in self.chunks:
            if live:
                expansion = unicode( eval( chunk, callerLocals, callerGlobals ) )
                if self.recursive:
                    expansion = unicode( kpyInterpol( expansion, self.recursive, self.defaultEncoding ) )
                expansions.append( expansion )
            else:
                expansions.append( unicode( chunk ) )
        try:
            R = u''.join( expansions )
        except UnicodeDecodeError:
            R = []
            for expansion in expansions:
                if isinstance( expansion, unicode ):
                    R.append( expansion )
                else:
                    R.append( expansion.decode( self.defaultEncoding ) )
            R = u''.join( R )
        return R

def getCallerNamespaces_original():
    try:
        try:
            1/0
        except:
            frame = sys.exc_traceback.tb_frame
        while frame.f_globals[ "__name__" ] == __name__:
            frame = frame.f_back
        callerLocals, callerGlobals = frame.f_locals, frame.f_globals
        return callerLocals, callerGlobals
    finally:
        del frame

def getCallerframe( framelevels = 2 ):
    """Return the frame that was valid ``levels`` calls away. The default for ``levels`` is ``framelevels``,
    which is equivalent to the frame of the code that called the code that called this function.

    OBS.: **Always** put your client code into a ``try``\/``finally`` block and **delete** the frame
    returned by this function. You will create circular references and, hence, memory leaks if you fail to
    delete the frame."""
    R = inspect.currentframe()
    for i in range( framelevels ):
        R = R.f_back
    return R

def getUpframe():
    """Get the first frame up in the call frame hierarchy where the value of the global variable
    ``__name__`` is not the same as ``__name__`` as valid within the namespace of this function."""
    R = inspect.currentframe()
    #raise '*********'
    #print '********'
    while R.f_globals[ "__name__" ] == __name__: R = R.f_back
    return R

def getCallerNamespaces( framelevels = 2 ):
    """Return the ``locals()`` and ``globals()`` of the code ``framelevels`` calls away. The default for
    ``framelevels`` is ``2``, which means you get back the namespaces of the code that called the code that
    called this function, which is the right thing if you use this function in your service function and
    want to look at the namespace of your client code."""
    try:
        if framelevels < 0:
            callerframe = getUpframe()
        else:
            callerframe = getCallerframe( framelevels + 1 )
        R = callerframe.f_locals, callerframe.f_globals
    finally:
        del callerframe
    return R



def inter( text ):

    return unicode( kpyInterpol( text ) )


if __name__ == '__main__':

    class X:

        DEFAULT_ENCODING = 'utf-8'

        value = u'ä'

        def __str__( self ):        return self.value.encode( self.DEFAULT_ENCODING )
        def __repr__( self ):       return repr( self.value )
        def __unicode__( self ):    return unicode( self.value )

    x = X()
    print `str( x )`
    print `repr( x )`
    print `unicode( x )`


    from kpyInterpol import kpyInterpol


    class testValues:
        class octets:
            umlauts0 = 'a'
            umlauts1 = 'ä'  # critical: this produces an octet string with the current encoding of this file
        class unicodes:
            umlauts0 = u'a'
            umlauts1 = u'ä'
    class templates:
        class octets:
            umlauts0o = '$testValues.octets.umlauts0'
            umlauts1o = '$testValues.octets.umlauts1'       # fails
            umlauts0u = '$testValues.unicodes.umlauts0'
            umlauts1u = '$testValues.unicodes.umlauts1'
        class unicodes:
            umlauts0o = u'$testValues.octets.umlauts0'
            umlauts1o = u'$testValues.octets.umlauts1'      # fails
            umlauts0u = u'$testValues.unicodes.umlauts0'
            umlauts1u = u'$testValues.unicodes.umlauts1'

    print '*****', `testValues.octets.umlauts1`

    def show( encoding, case, value ):
        print '-' * 108
        for r, n in zip( [ str, repr, unicode ], 'sru' ):
            print '%s%20s' % ( n, '%s.%s:' % ( encoding, case, ) ),
            try:
                #i = kpyInterpol( value, defaultEncoding = 'latin-1' )
                i = kpyInterpol( value, defaultEncoding = 'utf-8' )
                print `r( i )`
    #               print i,
    #               print `i`
            except Exception, e:
                print '(%s) %s' % ( e.__class__.__name__, e )

    for encoding in 'octets unicodes'.split():
        for case in 'umlauts0o umlauts1o umlauts0u umlauts1u'.split():
            #print '=' * 108
            #show( encoding, case, getattr( getattr( testValues, encoding ), case ), False )
            template = getattr( getattr(  templates, encoding ), case )
            show( encoding, case, template )
            #print kpyInterpol( template ).chunks
