#
# -*- coding: utf-8 -*-

"""Neo is a matrix and vector package specifically tailored to the needs of reprom. Methods include
fuzzy logic operators."""

############################################################################################################
#
#   YELLOW PAGES
#
#===========================================================================================================
from __future__ import                  division
#-----------------------------------------------------------------------------------------------------------
import                                  math
import                                  operator
#   import                                  Tkinter
#   import                                  colorsys
#-----------------------------------------------------------------------------------------------------------
u                                       = g.configtool.utilities
#-----------------------------------------------------------------------------------------------------------
SerializableMx                          = u.SerializableMx.SerializableMx
fuzzy1                                  = u.fuzzy.monopolar
fuzzy2                                  = u.fuzzy.bipolar
intfsFromfloatfs                        = u.typo.intfsFromfloatfs
#-----------------------------------------------------------------------------------------------------------
Misfit                                  = u.Misfit.Misfit
#-----------------------------------------------------------------------------------------------------------
_SEP0                                   = '#' * 108
_SEP1                                   = '=' * 108
_SEP2                                   = '-' * 108
VERBOSE                                 = False
#PHI                                     = ( math.sqrt( 5 ) - 1 ) / 2

############################################################################################################
#
#
#
#===========================================================================================================
class NeoError:
    pass

#===========================================================================================================
class NotImplementedError( NeoError, NotImplementedError ):

    #-------------------------------------------------------------------------------------------------------
    def f( this ):
        return this( "This method not implemented in base class." )
    notImplementedInBaseClass = classmethod( f ); del f

#===========================================================================================================
class VError( NeoError, ValueError ):

    #-------------------------------------------------------------------------------------------------------
    def f( this, sparse, vector ):
        vmin = sparse.vmin
        vmax = sparse.vmax
        return this(
            "Value exceeds permitted range: is %(vector)r, should be within %(vmin)s <= vector <= %(vmax)s."
                % vars() )
    valueNotInRange = classmethod( f ); del f

    #-------------------------------------------------------------------------------------------------------
    def f( this, thisvector, thatvector ):
        simin, simax = thisvector.imin, thisvector.imax
        oimin, oimax = thatvector.imin, thatvector.imax
        return this( "Vectors are incongruent: " +
            "this vector has index range (%(simin)s,%(simax)s), " % vars() +
            "other vector has index range (%(oimin)s,%(oimax)s)." % vars() )
    vectorIncongruent = classmethod( f ); del f

    #-------------------------------------------------------------------------------------------------------
    def f( this, thismatrix, thatmatrix ):
        simin, simax, sjmin, sjmax = thismatrix.imin, thismatrix.imax, thismatrix.jmin, thismatrix.jmax
        oimin, oimax, ojmin, ojmax = thatmatrix.imin, thatmatrix.imax, thatmatrix.jmin, thatmatrix.jmax
        return this( "Matrices are incongruent: " +
            "this matrix has index ranges (%(simin)s,%(simax)s) and (%(sjmin)s,%(sjmax)s), " % vars() +
            "other matrix has index ranges (%(oimin)s,%(oimax)s) and (%(ojmin)s,%(ojmax)s)." % vars() )
    matrixIncongruent = classmethod( f ); del f

#===========================================================================================================
class KError( NeoError, KeyError, IndexError ):

    #-------------------------------------------------------------------------------------------------------
    def f( this, sparse, key, kmin, kmax ):
        return this(
            "Index exceeds permitted range: is %(key)r, should be within %(kmin)s <= i < %(kmax)s."
                % vars() )
    integerIndexNotInRange = classmethod( f ); del f

    #-------------------------------------------------------------------------------------------------------
    def f( this, e ):
        return this( "Index must be a single integer: %(e)s." % vars() )
    indexMustBeInteger = classmethod( f ); del f

#===========================================================================================================
class TError( NeoError, TypeError ):

    #-------------------------------------------------------------------------------------------------------
    def f( this, minmax ):
        try:
            length = len( minmax )
        except:
            length = None
        return this(
            "Expected object of length two; has length %(length)s." % vars() )
    notAPair = classmethod( f ); del f

    #-------------------------------------------------------------------------------------------------------
    def f( this, vector ):
        vclass = vector.__class__
        return this( "Vector should be of class neo.Vector, is %(vclass)s." % vars() )
    notASparseVector = classmethod( f )

    #-------------------------------------------------------------------------------------------------------
    def f( this, vector ):
        vclass = vector.__class__
        return this( "Matrix should be of class neo.Matrix, is %(vclass)s." % vars() )
    notASparseMatrix = classmethod( f )

    #-------------------------------------------------------------------------------------------------------
    def f( this ):
        return this( "Vector is incongruent." % vars() )
    notACongruentVector = classmethod( f )

    #-------------------------------------------------------------------------------------------------------
    def f( this ):
        return this( "Matrix is incongruent." % vars() )
    notACongruentMatrix = classmethod( f )


############################################################################################################
#
#
#
#===========================================================================================================
class Sparse( SerializableMx, dict ):
    """Base class for sparse vectors and matrices; a derivative of ``dict``."""

    """``key``, the indices (``i`` in a vector, ``ij`` or ``(i,j)`` in a matrix)

    ``sk``, the 'sparse indices' in a ``Sparse`` object (those indices that are actually defined).

    ``ak``, all indices from the range of permitted indices.

    ``sv``, the 'sparse values' in a ``Sparse`` object (those values that are actually defined).

    ``ak``, all values from the elements within the range of permitted indices; undefined elements have the
    default value, ``~.vdefault``.

    ``sitems``, the 'sparse items' ``(sk,sv)``.


    """

    #-------------------------------------------------------------------------------------------------------
    _vdefault           = 0
    vmin                = 0
    vmax                = 1
    eachRoundDigits     = 8
    roundDigits         = 5

    #-------------------------------------------------------------------------------------------------------
    def __init__( self, vdefault = None, vminmax = None ):
        if vdefault is None:
            vdefault = self._vdefault
        self.vdefault = vdefault
        if vminmax is not None:
            self.assertIsPair( vminmax )
            self.vmin, self.vmax = vminmax
        self.extremum = None

    #-------------------------------------------------------------------------------------------------------
    def _copy( self, withValues = True ):
        R = self.__class__()
        R.vdefault          = self.vdefault
        R.vmin              = self.vmin
        R.vmax              = self.vmax
        R.eachRoundDigits   = self.eachRoundDigits
        R.roundDigits       = self.roundDigits
        R.extremum          = self.extremum
        if withValues:
            R.updateSparse( self )
        return R

    #-------------------------------------------------------------------------------------------------------
    def cleanup( self ):
        """Mandatory after ``~._vdefault`` has been changed to make sure all default values are taken off
        the record."""
        for key, value in self.iterAllItems():
            self[ key ] = value

    #-------------------------------------------------------------------------------------------------------
    def f():
        def fget( self ):
            return self._vdefault
        def fset( self, value ):
            self._vdefault = round( value, self.eachRoundDigits )
            self.cleanup()
        return locals()
    vdefault = property( **f() ); del f

    ########################################################################################################
    #   ITERATION
    #-------------------------------------------------------------------------------------------------------
#       def __iter__( self ):
#           """Same as ``~.iterSparseItems()``."""
#           return self.iterSparseItems()

    #-------------------------------------------------------------------------------------------------------
    def itervalues( self ):
        """Same as ``~.iterSparseValues()``."""
        return self.iterSparseValues()

    #-------------------------------------------------------------------------------------------------------
    def iterkeys( self ):
        """Same as ``~.iterSparseKeys()``."""
        return self.iterSparseKeys()

    #-------------------------------------------------------------------------------------------------------
    def iterValues( self ):
        """Same as ``~.iterSparseValues()``."""
        return self.iterSparseValues()

    #-------------------------------------------------------------------------------------------------------
    def iterKeys( self ):
        """Same as ``~.iterSparseKeys()``."""
        return self.iterSparseKeys()

    #-------------------------------------------------------------------------------------------------------
    def iterSparseItems( self ):
        """Iterate over pairs ``(key,w)`` (i.e. over sparse indices and sparse values)."""
        return super( Sparse, self ).iteritems()

    #-------------------------------------------------------------------------------------------------------
    def iterSparseKeys( self ):
        """Iterate over sparse keys."""
        return super( Sparse, self ).iterkeys()

    #-------------------------------------------------------------------------------------------------------
    def iterSparseValues( self ):
        """Iterate over sparse values."""
        return super( Sparse, self ).itervalues()

    #-------------------------------------------------------------------------------------------------------
    def iterAllKeys( self ):
        """Iterate over all possible keys. Raises ``NotImplementedError`` in base class."""
        raise NotImplementedError.notImplementedInBaseClass()

    #-------------------------------------------------------------------------------------------------------
    def iterAllItems( self ):
        """Iterate over all possible items; return ``~.vdefault`` for undefined elements. Must define
        ``~.iterAllKeys()`` to make this method work."""
        for k in self.iterAllKeys():
            yield k, self[ k ]

    #-------------------------------------------------------------------------------------------------------
    def iterAllValues( self ):
        """Iterate over the values of all possible keys; return ``~.vdefault`` for undefined elements. Must
        define ``~.iterAllKeys()`` to make this method work."""
        for k in self.iterAllKeys():
            yield self[ k ]

    #-------------------------------------------------------------------------------------------------------
    def iterAllItemsByDescendingValue( self ):
        sortlist = [ ( value, key ) for key, value in self.iterAllItems() ]
        sortlist.sort()
        sortlist.reverse()
        for value, key in sortlist:
            yield key, value



    #   def f():
    #       def fget( self ):
    #           return self.
    #       return locals()
    #   kmin = property

    ########################################################################################################
    #   ELEMENT ASSIGNMENT AND RETRIEVAL
    #-------------------------------------------------------------------------------------------------------
    def __getitem__( self, key ):
        """Retrieve a defined item, or ``~.vdefault`` for undefined ones. See ``~.assertLegalIndex()`` for
        possible errors."""
        key = self._correctKey( key )
        self.assertLegalIndex( key )
        R = self.get( key, Misfit )
        if R is not Misfit: return R
        return self.vdefault

    #-------------------------------------------------------------------------------------------------------
    def _correctKey( self, key ):
        """Derivatives that normalize, scale or otherwise alter keys of elements (like ``SymmetricMatrix``)
        override this method to define the transformation."""
        return key

    #-------------------------------------------------------------------------------------------------------
    def __setitem__( self, key, value ):
        key = self._correctKey( key )
        self.assertLegalIndex( key )
        value = round( value, self.eachRoundDigits )
        if value == self.vdefault:
            if key in self:
                del self[ key ]
            return
        super( Sparse, self ).__setitem__( key, value )

    #-------------------------------------------------------------------------------------------------------
    def updateAll( self, other ):
        """Update *all* elements from the other ``Sparse`` object.

        This method should be preferred whenever it is not explicitly desired to only overwrite thos
        elements in ``self`` with elements from ``other`` that are actually defined in ``other``. Probably
        most of the time, we will think of sparse vectors and matrices as array-like objects that are,
        albeit 'sparingly defined', still compact arrays of values; their overall integrity (meaningfulness)
        is *not* solely determined by those values that differ from the default value, but by *all* the
        possible values (at the right places). For this reason, the original ``~.update()`` has been renamed
        to ``~.updateSparse()`` and its place taken by ``updateAll()``."""
        self.assertHasMyClassAndIsCongruent( other )
        for key, value in other.iterAllItems():
            self[ key ] = value

    #-------------------------------------------------------------------------------------------------------
    def update( self, other ):
        """Update *all* elements from the other ``Sparse`` object. Synonymous to ``~.updateAll()``, q.v. Use
        ``~.updateSparse()`` to only update elements whose value in ``other`` deviates from
        ``other.vdefault``."""
        self.updateAll( other )

    #-------------------------------------------------------------------------------------------------------
    def updateSparse( self, other ):
        """Update those elements in ``self`` whose counterpart in ``other`` deviates from
        ``other.vdefault``."""
        super( Sparse, self ).update( other )

    #-------------------------------------------------------------------------------------------------------
    def filter( self, f ):
        """Delete all elements from object for which ``f`` is not true."""
        assert callable( f ), "Filter must be callable"
        for key, value in self.iterAllItems():
            if not f( value ):
                del self[ key ]

    ########################################################################################################
    #   ARITHMETICS
    #-------------------------------------------------------------------------------------------------------
    def descale( self ):
        """This is a kind of normalizing method; it looks for the value in the matrix or vector with the
        largest absolute value and devides all values in the instance by that factor. That extremum is,
        appropriately, kept as ``~.extremum`` to make the transformation reversible. If the extremum is
        found to be zero, the extremum will be one (to avoid division by zero). NB. that this method does
        not normalize the object in the sense that its cartesian length becomes one or it is perpendicular
        to some other entity; to avoid confusion, the method's name reminds of 'de-escalation' rather than
        'normalization'."""
        #   minimum = +1e12
        #   maximum = -1e12
        values = self.values()
        if not values:
            values = [ self.vdefault, ]
        #   minimum = min( min( values ), minimum )
        #   maximum = max( max( values ), maximum )
        minimum = min( values )
        maximum = max( values )
        if abs( minimum ) >= abs( maximum ):
            extremum = abs( minimum )
        else:
            extremum = maximum
        if extremum == 0:
            extremum = 1
        self.vmin = self.vmax = None
        self.divS( extremum )
        self.vmin, self.vmax = -1, +1
        self.extremum = extremum

    #-------------------------------------------------------------------------------------------------------
    def squeeze( self ):
        values = self.values()
        if not values:
            values = [ self.vdefault, ]
        minimum = min( values )
        maximum = max( values )
        delta = abs( maximum - minimum ) / 2
        if delta == 0: delta = 1
        newVdefault = ( self.vdefault - minimum ) / delta
        self.vmin = self.vmax = None
        self.subS( minimum )
        self.divS( delta )
        self.subS( 1 )
        self.vmin, self.vmax = -1, +1
        self.vdefault = newVdefault

    #-------------------------------------------------------------------------------------------------------
    def round( self ):
        digits = self.roundDigits
        for key, value in self.iterSparseItems():
            self[ key ] = round( value, digits )

#       def __contains__( self, key ):
#           pass

    #-------------------------------------------------------------------------------------------------------
    #   PROPRIOCEPTIVE ELEMENT-WISE METHODS
    #-------------------------------------------------------------------------------------------------------
    def _doUnaryElementwiseOperation( self, op ):
        for k in self.iterAllKeys():
            self[ k ] = op( self[ k ] )

    #-------------------------------------------------------------------------------------------------------
    def _doBinaryElementwiseOperation( self, vectorOrMatrix, op ):
        self.assertHasMyClassAndIsCongruent( vectorOrMatrix )
        for k, value in vectorOrMatrix.iterAllItems():
            self[ k ] = op( self[ k ], value )

    #-------------------------------------------------------------------------------------------------------
    def meanE( self, vectorOrMatrix ):
        self._doBinaryElementwiseOperation( vectorOrMatrix, lambda x, y: ( x + y ) / 2 )

    #-------------------------------------------------------------------------------------------------------
    def addE( self, vectorOrMatrix ):
        self._doBinaryElementwiseOperation( vectorOrMatrix, operator.add )

    #-------------------------------------------------------------------------------------------------------
    def subE( self, vectorOrMatrix ):
        self._doBinaryElementwiseOperation( vectorOrMatrix, operator.sub )

    #-------------------------------------------------------------------------------------------------------
    def mulE( self, vectorOrMatrix ):
        self._doBinaryElementwiseOperation( vectorOrMatrix, operator.mul )

    #-------------------------------------------------------------------------------------------------------
    def divE( self, vectorOrMatrix ):
        self._doBinaryElementwiseOperation( vectorOrMatrix, operator.truediv )

    #-------------------------------------------------------------------------------------------------------
    def fuzzy1andE( self, vectorOrMatrix ):
        self._doBinaryElementwiseOperation( vectorOrMatrix, fuzzy1.AND )

    #-------------------------------------------------------------------------------------------------------
    def fuzzy1orE( self, vectorOrMatrix ):
        self._doBinaryElementwiseOperation( vectorOrMatrix, fuzzy1.OR )

    #-------------------------------------------------------------------------------------------------------
    def fuzzy1notE( self ):
        self._doUnaryElementwiseOperation( fuzzy1.NOT )

    #-------------------------------------------------------------------------------------------------------
    def fuzzy2andE( self, vectorOrMatrix ):
        self._doBinaryElementwiseOperation( vectorOrMatrix, fuzzy2.AND )

    #-------------------------------------------------------------------------------------------------------
    def fuzzy2orE( self, vectorOrMatrix ):
        self._doBinaryElementwiseOperation( vectorOrMatrix, fuzzy2.OR )

    #-------------------------------------------------------------------------------------------------------
    def fuzzy2notE( self ):
        self._doUnaryElementwiseOperation( fuzzy2.NOT )

    #-------------------------------------------------------------------------------------------------------
    def fuzzyAndE( self, vectorOrMatrix ):
        if self.vmin == 0 and self.vmax == 1:
            return self._doBinaryElementwiseOperation( vectorOrMatrix, fuzzy1.AND )
        elif self.vmin == -1 and self.vmax == 1:
            return self._doBinaryElementwiseOperation( vectorOrMatrix, fuzzy2.AND )
        raise TypeError( "Can only use this method if both vmin and vmax attributes set correctly." )

    #-------------------------------------------------------------------------------------------------------
    def fuzzyOrE( self, vectorOrMatrix ):
        if self.vmin == 0 and self.vmax == 1:
            return self._doBinaryElementwiseOperation( vectorOrMatrix, fuzzy1.OR )
        elif self.vmin == -1 and self.vmax == 1:
            return self._doBinaryElementwiseOperation( vectorOrMatrix, fuzzy2.OR )
        raise TypeError( "Can only use this method if both vmin and vmax attributes set correctly." )

    #-------------------------------------------------------------------------------------------------------
    def fuzzyNotE( self ):
        if self.vmin == 0 and self.vmax == 1:
            return self._doUnaryElementwiseOperation( fuzzy1.NOT )
        elif self.vmin == -1 and self.vmax == 1:
            return self._doUnaryElementwiseOperation( fuzzy2.NOT )
        raise TypeError( "Can only use this method if both vmin and vmax attributes set correctly." )

    #-------------------------------------------------------------------------------------------------------
    #   PROPRIOCEPTIVE SCALAR METHODS
    #-------------------------------------------------------------------------------------------------------
    def _doBinaryScalarOperation( self, scalar, op ):
        self.assertIsScalar( scalar )
        for k in self.iterAllKeys():
            self[ k ] = op( self[ k ], scalar )

    #-------------------------------------------------------------------------------------------------------
    def addS( self, scalar ):
        self._doBinaryScalarOperation( scalar, operator.add )

    #-------------------------------------------------------------------------------------------------------
    def subS( self, scalar ):
        self._doBinaryScalarOperation( scalar, operator.sub )

    #-------------------------------------------------------------------------------------------------------
    def mulS( self, scalar ):
        self._doBinaryScalarOperation( scalar, operator.mul )

    #-------------------------------------------------------------------------------------------------------
    def divS( self, scalar ):
        self._doBinaryScalarOperation( scalar, operator.truediv )

    #-------------------------------------------------------------------------------------------------------
    def fuzzy1andS( self, vectorOrMatrix ):
        self._doBinaryScalarOperation( vectorOrMatrix, fuzzy1.AND )

    #-------------------------------------------------------------------------------------------------------
    def fuzzy1orS( self, vectorOrMatrix ):
        self._doBinaryScalarOperation( vectorOrMatrix, fuzzy1.OR )

    #-------------------------------------------------------------------------------------------------------
    def fuzzy2andS( self, vectorOrMatrix ):
        self._doBinaryScalarOperation( vectorOrMatrix, fuzzy2.AND )

    #-------------------------------------------------------------------------------------------------------
    def fuzzy2orS( self, vectorOrMatrix ):
        self._doBinaryScalarOperation( vectorOrMatrix, fuzzy2.OR )

#       ########################################################################################################
#       #   METHOD SELECTORS
#       #-------------------------------------------------------------------------------------------------------
#       def _selectElementwiseOrScalarMethod( opbasename, scalarOrSparse ):
#           if isinstance( scalarOrSparse, Sparse ):
#               opname = '%sE' % ( opbasename, )
#           else:
#               opname = '%sS' % ( opbasename, )
#           return getattr( self, opname )

    ########################################################################################################
    #   TESTS AND ASSERTIONS
    #-------------------------------------------------------------------------------------------------------
    def isPossibleI( self, i ):
        return i in self.iterAllI()

    #-------------------------------------------------------------------------------------------------------
    def assertLegalIndex( self, key ):
        """In derivatives, raises a ``neo.KError`` (which is both a ``KeyError`` and an ``IndexError``) if
        index ``key`` is outside the legal indices; in base class, raises ``NotImplementedError``."""
        raise NotImplementedError.notImplementedInBaseClass()

    #-------------------------------------------------------------------------------------------------------
    def assertScalarIndexWithinRange( self, k, kmin, kmax ):
        """Raises a ``neo.KError`` if index ``k`` is outside the legal indices."""
        if kmin is not None and k < kmin:
            raise KError.integerIndexNotInRange( self, k, kmin, kmax )
        if kmax is not None and k >= kmax:
            raise KError.integerIndexNotInRange( self, k, kmin, kmax )

    #-------------------------------------------------------------------------------------------------------
    def assertLegalValue( self, vector ):
        """Raises a ``neoVError`` (derived from ``ValueError``) if ``vector`` is outside the range of legal
        values."""
        if self.vmin is not None:
            if vector < self.vmin:
                raise VError.valueNotInRange( self, vector )
        if self.vmax is not None:
            if vector > self.vmax:
                raise VError.valueNotInRange( self, vector )

    #-------------------------------------------------------------------------------------------------------
    def assertIsPair( self, x ):
        if not len( x ) == 2:
            raise TError.notAPair( x )

    #-------------------------------------------------------------------------------------------------------
    def assertIsInteger( self, minmax ):
        raise NotImplementedError.notImplementedInBaseClass()

    #-------------------------------------------------------------------------------------------------------
    def assertIsSparseVector( self, x ):
        if not isinstance( x, Vector ):
            raise TError.notASparseVector( x )

    #-------------------------------------------------------------------------------------------------------
    def assertIsSparseMatrix( self, x ):
        if not isinstance( x, Matrix ):
            raise TError.notASparseMatrix( x )

#       #-------------------------------------------------------------------------------------------------------
#       def assertIsCongruent( self, vectorOrMatrix ):
#           """In derivatives, raises a ``neo.TError`` if ``vectorOrMatrix`` does not have the same minimal and
#           maximal coordinates; in base class, raises ``NotImplementedError``."""
#           raise NotImplementedError.notImplementedInBaseClass()

    #-------------------------------------------------------------------------------------------------------
    def assertIsScalar( self, scalar ):
        pass
        #   if not scalar == float( scalar ):
        #       raise

    #-------------------------------------------------------------------------------------------------------
    def assertIsInteger( self, x ):
        """Raises a ``neo.KError`` if ``x`` is not an integer."""
        try:
            if not x == int( x ):
                raise ValueError( "Must be integer, is %r." % ( x, ) )
        except Exception, e:
            raise KError.indexMustBeInteger( e )

#   #-------------------------------------------------------------------------------------------------------
#   def assertIsReal( self, x ):
#       """Raises a ``neo.KError`` if ``x`` is not an integer."""
#       try:
#           if not x == float( x ):
#               raise ValueError( "Must be integer, long or float, is %r." % ( x, ) )
#       except Exception, e:
#           raise KError.indexMustBeInteger( e )

    ########################################################################################################
    #   REPRESENTATION
    #-------------------------------------------------------------------------------------------------------
    def getstr( self, d0 = 3, d1 = 0, title = True, separator = True, values = True ):
        return '\n'.join( self.getstrs( d0, d1, title, separator, values ) )

    #-------------------------------------------------------------------------------------------------------
    def getstrs( self, d0 = 3, d1 = 0, title = True, separator = True, values = True ):
        floatfs, intfs = self._getFormatstrings( d0, d1 )
        R = []
        if values:
            blank = ' ' * len( floatfs % 1 )
            v = []
            for x in self.iterAllValues():
                if x == self.vdefault:
                    v.append( blank )
                else:
                    v.append( floatfs % x  )
            R.append( ' '.join( v ) )
        if title:
            v_ = ' '.join( [ intfs % i for i in self.iterAllKeys() ] )
            if separator:
                R.append( '-' * len( v_ ) )
            R.append( v_ )
        return R

    #-------------------------------------------------------------------------------------------------------
    def f( this, d0, d1 ):
        floatfs = '%%+%d.%df' % ( d0, d1 )
        intfs = intfsFromfloatfs( floatfs )
        return floatfs, intfs
    _getFormatstrings = classmethod( f ); del f


############################################################################################################
#
#   AFTERTHOUGHTS
#
#===========================================================================================================

#-----------------------------------------------------------------------------------------------------------
#   REFLECTIVE METHODS
#-----------------------------------------------------------------------------------------------------------
def _setReflectiveFromProprioceptiveMethod( pname ):
    global Sparse
    def getwrap( unboundProprioceptiveMethod ):
        def wrap( self, *P, **Q ):
            R = self.copy()
            unboundProprioceptiveMethod( R, *P, **Q )
            return R
        return wrap
    rname = 'r%s%s' % ( pname[ 0 ].upper(), pname[ 1 : ], )
    setattr( Sparse, rname, getwrap( getattr( Sparse, pname ) ) )

for pname in """
    addE subE mulE divE addS subS mulS divS
    fuzzy1orS fuzzy2orS fuzzy1andS fuzzy2andS
    fuzzy1orE fuzzy2orE fuzzy1andE fuzzy2andE
    """.split():
    _setReflectiveFromProprioceptiveMethod( pname )
#
#
#   _selectElementwiseOrScalarMethod( opbasename, scalarOrSparse )#################################

#-----------------------------------------------------------------------------------------------------------
#   MAGIC METHODS
#-----------------------------------------------------------------------------------------------------------
Sparse.__add__      = Sparse.__iadd__       = Sparse.rAddE
Sparse.__sub__      = Sparse.__isub__       = Sparse.rSubE
Sparse.__mul__      = Sparse.__imul__       = Sparse.rMulE
Sparse.__div__      = Sparse.__idiv__       = Sparse.rDivE
Sparse.__truediv__  = Sparse.__itruediv__   = Sparse.rDivE

############################################################################################################
#
#
#
#===========================================================================================================
class Vector( Sparse ):

    #-------------------------------------------------------------------------------------------------------
    imin = 0
    imax = 0

    #-------------------------------------------------------------------------------------------------------
    def __init__( self, iminmax = None, vminmax = None ):
        super( Vector, self ).__init__( vminmax )
        if iminmax is not None:
            self.assertIsPair( iminmax )
            self.imin, self.imax = iminmax

    #-------------------------------------------------------------------------------------------------------
    def copy( self, withValues = True ):
        R = super( Vector, self )._copy( withValues )
        R.imin      = self.imin
        R.imax      = self.imax
        return R

    ########################################################################################################
    #   ITERATION
    #-------------------------------------------------------------------------------------------------------
    def rangeOfAllI( self ):
        return xrange( self.imin, self.imax )

    #-------------------------------------------------------------------------------------------------------
    def iterAllKeys( self ):
        """Iterate over all possible keys."""
        for i in self.rangeOfAllI():
            yield i

    ########################################################################################################
    #   TRANSFORMATION
    #-------------------------------------------------------------------------------------------------------
    def spanMatrix( self, other, f = None, vdefault = None ):
        #---------------------------------------------------------------------------------------------------
        self.assertIsSparseVector( other )
        #---------------------------------------------------------------------------------------------------
        if f is None: f = lambda x, y: x + y
        if vdefault is None: vdefault = self.vdefault
        #---------------------------------------------------------------------------------------------------
        R = Matrix(
            iminmax     = (  self.imin,  self.imax ),
            jminmax     = ( other.imin, other.imax ),
            vdefault    = vdefault,
            vminmax     = (  self.vmin,  self.vmax ),
            )
        for i, x in self.iterAllItems():
            for j, y in other.iterAllItems():
                R[ i,j ] = f( x, y )
        return R

    #-------------------------------------------------------------------------------------------------------
    def spanSymmetricMatrix( self, f = None ):
        return self.spanMatrix( self, f )


    ########################################################################################################
    #   ARITHMETICS

    ########################################################################################################
    #   TESTS AND ASSERTIONS
    #-------------------------------------------------------------------------------------------------------
    def assertLegalIndex( self, i ):
        """Raises a ``neo.KError`` if ``i`` is not an integer within the range of legal indices."""
        self.assertIsInteger( i )
        self.assertScalarIndexWithinRange( i, self.imin, self.imax  )

#       #-------------------------------------------------------------------------------------------------------
#       def assertScalarIndexWithinRange( self, i ):
#           """Raises a ``neo.KError`` if index ``i`` is outside the legal indices."""
#           imin, imax = self.imin, self.imax
#           if imin is not None:
#               if i < imin:
#                   raise KError.integerIndexNotInRange( self, i, imin, imax )
#           if imax is not None:
#               if i >= imax:
#                   raise KError.integerIndexNotInRange( self, i, imin, imax )

    #-------------------------------------------------------------------------------------------------------
    def assertIsCongruent( self, vector ):
        """Raise a ``neo.TError`` if ``vector`` does not have the same minimal and maximal coordinates."""
        if not self.imin == vector.imin and self.imax == vector.imax:
            raise VError.vectorIncongruent( self, vector )

    #-------------------------------------------------------------------------------------------------------
    def assertHasMyClassAndIsCongruent( self, vector ):
        self.assertIsSparseVector( vector )
        self.assertIsCongruent( vector )


############################################################################################################
#
#
#
#===========================================================================================================
class Matrix( Sparse ):

    #-------------------------------------------------------------------------------------------------------
    imin = 0
    imax = 0
    jmin = 0
    jmax = 0

    #-------------------------------------------------------------------------------------------------------
    def __init__( self, iminmax = None, jminmax = None, vdefault = None, vminmax = None ):
        super( Matrix, self ).__init__( vdefault, vminmax )
        if iminmax is not None:
            self.assertIsPair( iminmax )
            self.imin, self.imax = iminmax
        if jminmax is not None:
            self.assertIsPair( jminmax )
            self.jmin, self.jmax = jminmax

    #-------------------------------------------------------------------------------------------------------
    def copy( self, withValues = True ):
        R = super( Matrix, self )._copy( withValues )
        R.imin      = self.imin
        R.imax      = self.imax
        R.jmin      = self.jmin
        R.jmax      = self.jmax
        return R

    #-------------------------------------------------------------------------------------------------------
    def f():
        def fget( self ):
            return self.imax - self.imin
        return locals()
    leni = property( **f() ); del f

    #-------------------------------------------------------------------------------------------------------
    def f():
        def fget( self ):
            return self.jmax - self.jmin
        return locals()
    lenj = property( **f() ); del f

    #-------------------------------------------------------------------------------------------------------
    def f():
        def fget( self ):
            return self.leni * self.lenj
        return locals()
    size = property( **f() ); del f

    ########################################################################################################
    #   ITERATION
    #-------------------------------------------------------------------------------------------------------
    def rangeOfAllI( self ):
        return xrange( self.imin, self.imax )

    #-------------------------------------------------------------------------------------------------------
    def rangeOfAllJ( self ):
        return xrange( self.jmin, self.jmax )

    #-------------------------------------------------------------------------------------------------------
    def iterAllKeys( self ):
        """Iterate over all possible keys."""
        for i in self.rangeOfAllI():
            for j in self.rangeOfAllJ():
                yield i,j

    #-------------------------------------------------------------------------------------------------------
    def iterAllI( self ):
        for i in self.rangeOfAllI():
            yield i

    #-------------------------------------------------------------------------------------------------------
    def iterAllJ( self ):
        for j in self.rangeOfAllJ():
            yield j

    ########################################################################################################
    #   EXTRACTING VECTORS
    #-------------------------------------------------------------------------------------------------------
    def getColumnVector( self, i ):
        R           = Vector()
        R.imin      = self.jmin
        R.imax      = self.jmax
        R.vdefault  = self.vdefault
        for j in self.iterAllJ():
            R[ j ]  = self[ i,j ]
        return R

    #-------------------------------------------------------------------------------------------------------
    def getRowVector( self, j ):
        R           = Vector()
        R.imin      = self.imin
        R.imax      = self.imax
        R.vdefault  = self.vdefault
        for i in self.iterAllI():
            R[ i ]  = self[ i,j ]
        return R

    ########################################################################################################
    #   TESTS AND ASSERTIONS
    #-------------------------------------------------------------------------------------------------------
    def isPossibleJ( self, j ):
        return j in self.iterAllJ()

    #-------------------------------------------------------------------------------------------------------
    def assertLegalIndex( self, ij ):
        """Raises a ``neo.KError`` if ``ij`` is not a pair of integers within the range of legal indices."""
        self.assertIsPair( ij )
        i,j = ij
        self.assertIsInteger( i )
        self.assertIsInteger( j )
        self.assertScalarIndexWithinRange( i, self.imin, self.imax )
        self.assertScalarIndexWithinRange( j, self.jmin, self.jmax )

    #-------------------------------------------------------------------------------------------------------
    def assertIsCongruent( self, matrix ):
        """Raise a ``neo.TError`` if ``matrix`` does not have the same minimal and maximal coordinates."""
        if not  self.imin == matrix.imin and self.imax == matrix.imax \
            and self.jmin == matrix.jmin and self.jmax == matrix.jmax:
            raise VError.matrixIncongruent( self, matrix )

    #-------------------------------------------------------------------------------------------------------
    def assertHasMyClassAndIsCongruent( self, matrix ):
        self.assertIsSparseMatrix( matrix )
        self.assertIsCongruent( matrix )

    #-------------------------------------------------------------------------------------------------------
    def getstr( self, d0 = 3, d1 = 0 ):
        imin, imax = self.imin, self.imax
        floatfs, intfs = Vector._getFormatstrings( d0, d1 )
        R = []
        js = list( self.iterAllJ() )
        js.reverse()
        for j in js:
            rowVector = self.getRowVector( j = j ).getstr( d0, d1, title = False )
            R.append( '%s|%s' % ( intfs % j, rowVector ) )
        v = Vector( iminmax = ( self.imin, self.imax ) )
        for i in self.iterAllI(): v[ i ] = i
        separator, title = v.getstrs( d0, d1, title = True, separator = True, values = False )
        #print '#########separator', separator
        emptyspace = ( ' ' * len( floatfs % 0 ) ) + '|'
        presep     = ( '-' * len( floatfs % 0 ) ) + '|'
        R.append( '%s%s' % ( presep,     separator ) )
        R.append( '%s%s' % ( emptyspace, title ) )
        return '\n'.join( R )


############################################################################################################
#
#
#
#===========================================================================================================
class SymmetricMatrix( Matrix ):

    #-------------------------------------------------------------------------------------------------------
    def __init__( self, iminmax = None, vdefault = None, vminmax = None ):
        super( SymmetricMatrix, self ).__init__( iminmax, iminmax, vdefault, vminmax )

    #-------------------------------------------------------------------------------------------------------
    def _correctKey( self, key ):
        """Derivatives that normalize, scale or otherwise alter keys of elements (like ``SymmetricMatrix``)
        override this method to define the transformation."""
        i,j = key
        if i < j: i,j = j,i
        return i,j


############################################################################################################
#
#
#
#===========================================================================================================
if __name__ == '__main__':

    import sys; sys.setrecursionlimit(50)


    v0 = Vector()
    v0.eachRoundDigits = 3
    v0.imax = 5
    v0.imin = -5
    v0[ 2 ] = 4
    print v0
    v0.addS( 3 )
    print v0
    v0.subS( 3 )
    print v0
    v0.addS( 3 )
    v0.divS( 1.5 )
    print v0
    print round( v0[ 2 ], 5 )
    #print v0[ 2 ], v0[ 2 ] == 4.7
    print

    v1 = Vector()
    v1.eachRoundDigits = 3
    v1.imax = 5
    v1.imin = -5
    v1[ 1 ] = 30
    v0.addE( v1 )
    print v0
    print

    v2 = v0.copy()
    print v2
    print id( v0 ), id( v2 )
    print v0.__dict__, v1.__dict__
    print v0.getstr( 6, 2 )
    print

    print _SEP0
    print 'Bipolar Fuzzy Operators'
    mi = Matrix( iminmax = ( -4, 5 ), jminmax = ( -4, 5 ), vdefault = 0, vminmax = ( -1, 1 ) )
    for i in mi.iterAllI():
        value = i / 4
        for j in mi.iterAllJ():
            mi[ i,j ] = value

    mj = Matrix( iminmax = ( -4, 5 ), jminmax = ( -4, 5 ), vdefault = 0, vminmax = ( -1, 1 ) )
    for j in mj.iterAllJ():
        value = j / 4
        for i in mj.iterAllI():
            mj[ i,j ] = value

    print _SEP0
    print mi.getstr( 5, 2 )
    print
    print mj.getstr( 5, 2 )
    print

    print _SEP1
    mi.fuzzy2notE()
    #mi.fuzzy2andE( mj )
    #mi.fuzzy2orS( .75 )
    print mi.getstr( 5, 2 )

    print _SEP0
    print 'Monopolar Fuzzy Operators'
    mi = Matrix( iminmax = ( 0, 5 ), jminmax = ( 0, 5 ), vdefault = 0, vminmax = ( 0, 1 ) )
    for i in mi.iterAllI():
        value = i / 4
        for j in mi.iterAllJ():
            mi[ i,j ] = value

    mj = Matrix( iminmax = ( 0, 5 ), jminmax = ( 0, 5 ), vdefault = 0, vminmax = ( 0, 1 ) )
    for j in mj.iterAllJ():
        value = j / 4
        for i in mj.iterAllI():
            mj[ i,j ] = value

    print _SEP0
    print mi.getstr( 5, 2 )
    print
    print mj.getstr( 5, 2 )
    print

    print _SEP1
    #mi.fuzzy1notE()
    #mi.fuzzy1andE( mj )
    mi.fuzzy1orE( mj )
    #mi.fuzzy1orS( .75 )
    print mi.getstr( 5, 2 )
    print

    #mk = mi.rAddE( mj )
    mk = mi + mj
    mk += mi
    print id( mk ), id( mi ), id( mj )
    print mk.getstr( 5, 2 )

    print
    print _SEP0
    m0 = SymmetricMatrix( ( 0, 10 ) )
    m0[ 2,2 ] = 22
    m0[ 2,3 ] = 23
    m0[ 4,3 ] = 43
    m0[ 5,3 ] = 53
    m0[ 6,3 ] = 63
    print m0.getstr( 5, 0 )

    print
    print _SEP0
    v0 = Vector( (  0, 10 ) )
    v1 = Vector( ( 10, 20 ) )
    for i in v0.iterAllKeys():
        v0[ i ] = i
    for j in v1.iterAllKeys():
        v1[ j ] = j
    m = v0.spanMatrix( v1 )
    print m.getstr( 4 )

    print
    print _SEP0
    m1 = m.copy()
    m1.mulS( 2 )
    m.meanE( m1 )
    print m.getstr( 4 )

    print ' '.join( dir( m ) )

    v = Vector( iminmax = ( 0, 5 ) )
    for i, x in enumerate( ( 0.2, -1 ) ):
        v[ i ] = x
    print v
    v.squeeze()
    print v


#       m0.mulS( 0.5 )
#       print
#       print m0.getstr( 7, 3 )
#
#       m0.addS( 0.5 )
#       m0.vdefault = 0.5
#       print
#       print m0.getstr( 7, 3 )

#       m0.mulS( 0.9 )
#       print
#       print m0.getstr( 5, 3 )
#
#       m0.addS( 0.5 )
#       print
#       print m0.getstr( 5, 3 )
#
#       m0.vdefault = 0.05
#       print
#       print m0.getstr( 5, 3 )
#
#
#



#       print
#       print mi
#       print
#       print mj.getstr( 5, 2 )
#       print

#       m0[ 7,28 ] = 1
#       print m0.getstr()
#       print
#
#       m1 = m0.copy()
#       m1[ 0, 20 ] = 42
#       m1.addS( 0.2 )
#       m0.addE( m1 )
#       print m0.getstr()
#       print
#
#       print m1.getstr( 6, 3 )
#       print
#
#       m1.fuzzy1orS( 0.3 )
#       print m1.getstr( 6, 3 )
#       print
#
#       m1.fuzzy1notE()
#       print m1.getstr( 6, 3 )
#       print





