#
# -*- coding: utf-8 -*-

"""

Tokenizer class.

To do
=====

*   Class ``PureCamelCaseTokenizer`` is currently derived from ``CamelCaseTokenizer``, which is historically
correct but crazy otherwise. Also, there should be a better way to derive tokenizer classes; since there are
at lot of method calls in ``~.tokenize()``, modified derivations are currently not so easily done.


"""

############################################################################################################
#
#   YELLOW PAGES
#
#===========================================================================================================
import re
from sets import Set
#-----------------------------------------------------------------------------------------------------------
u                                       = g.configtool.utilities
#-----------------------------------------------------------------------------------------------------------
nlnormalizer                            = u.nlnormalizer
#-----------------------------------------------------------------------------------------------------------
stopwords                               = nlnormalizer.stopwords_general_en.stopwords
characterMappings                       = nlnormalizer.LatinNormalizer.LatinNormalizer.fromComplexToPlain
#-----------------------------------------------------------------------------------------------------------
DEBUG                                   = False
#   DEBUG                                   = True
#-----------------------------------------------------------------------------------------------------------


############################################################################################################
#
#   TOKENIZER CLASS
#
#===========================================================================================================
class Tokenizer:

    #=======================================================================================================
    LOWER                           = 'lower'
    UPPER                           = 'upper'
    #-------------------------------------------------------------------------------------------------------
    minimalWordLength               = 3                                         # an int
    stopwords                       = stopwords                                 # a Set
    illegalCharactersReAlpha        = re.compile( '[^a-zA-Z]' )                 # a regex
    #illegalCharactersReAlphaSigned  = re.compile( '[^a-zA-Z]|(^[+-])|(?<=[ \t])' )               # a regex
    illegalCharactersReAlphaSigned  = re.compile( '[^a-zA-Z+-]' )               # a regex
    illegalCharactersReAlphanum     = re.compile( '[^a-zA-Z0-9]' )              # a regex
    illegalCharactersSubstitute     = ' '                                       # a string
    characterMappings               = characterMappings                         # a dict
    normalCase                      = LOWER                                     # UPPER or LOWER
    postpurgativeTextCorrections    = []                                        # a list of pairs
    prepurgativeTextCorrections     = []                                        # a list of pairs
    tokenCorrections                = []                                        # a list of pairs
    shortTerms                      = Set()                                     # a Set
    allowSigns                      = False
    allowNumbers                    = False

    #-------------------------------------------------------------------------------------------------------
    def __init__( self, **Q ):
        """Users are free to pass any keyword arguments at initialization time which will become instance
        attributes and override homonymous class attributes."""
        self.__dict__.update( Q )

    ########################################################################################################
    #   MAIN METHOD
    #=======================================================================================================
    def tokenize( self, text ):#, _debug = False ):
        #---------------------------------------------------------------------------------------------------
        text = self.normalizeAccentedCharacters( text )
        text = self.doPrepurgativeTextCorrections( text )
        text = self.purgeIllegalCharacters( text )
        text = self.doPostpurgativeTextCorrections( text )
        text = self.normalizeCase( text )
        #---------------------------------------------------------------------------------------------------
        tokens = self.segment( text )
        #---------------------------------------------------------------------------------------------------
        tokens = self.doTokenCorrections( tokens )#, _debug )
        tokens = self.removeShortWords( tokens )
        tokens = self.removeStopwords( tokens )
        tokens = self.stem( tokens )
        #---------------------------------------------------------------------------------------------------
        return tokens

    ########################################################################################################
    #   METHODS ACTING ON TEXT
    #=======================================================================================================
    def purgeIllegalCharacters( self, text ):
        #   if self.returnNumbersAsNumbers: illegalCharactersRe = self.illegalCharactersReAlphanum
        if self.allowSigns:
            illegalCharactersRe = self.illegalCharactersReAlphaSigned
        else:
            illegalCharactersRe = self.illegalCharactersReAlpha
        return illegalCharactersRe.sub( self.illegalCharactersSubstitute, text )

    #=======================================================================================================
    def normalizeAccentedCharacters( self, text ):
        tt = self.characterMappings
        return ''.join( [ tt.get( c, c ) for c in text ] )

    #=======================================================================================================
    def normalizeCase( self, text ):
        return getattr( text, self.normalCase )()

    #=======================================================================================================
    def doPrepurgativeTextCorrections( self, text ):
        return self._doTextCorrections( text, self.prepurgativeTextCorrections )
    #-------------------------------------------------------------------------------------------------------
    def doPostpurgativeTextCorrections( self, text ):
        return self._doTextCorrections( text, self.postpurgativeTextCorrections )
    #-------------------------------------------------------------------------------------------------------
    def _doTextCorrections( self, text, textCorrections ):
        R = text
        for regex, substitution in textCorrections:
            if DEBUG and ( regex.search( R ) is not None ):
                print '###', repr( regex.pattern ), regex.sub( r'[\g<0>=>%s]' % ( substitution, ), R )
            R = regex.sub( substitution, R )
        return R

    ########################################################################################################
    #   METHOD TO TURN TEXT INTO TOKENS
    #=======================================================================================================
    def segment( self, text ):
        return text.split()

    ########################################################################################################
    #   METHODS ACTING ON TOKENS
    #=======================================================================================================
    def removeStopwords( self, tokens ):
        sw = self.stopwords
        return [ term for term in tokens if term not in sw ]

    #=======================================================================================================
    def removeShortWords( self, tokens ):
        minlen      = self.minimalWordLength
        shortTerms  = self.shortTerms
        return [ token for token in tokens if ( len( token ) >= minlen ) or ( token in shortTerms ) ]

    #=======================================================================================================
    def doTokenCorrections( self, tokens ):#, _debug = False ):
        tokenCorrections = self.tokenCorrections
        for wrong, correct in tokenCorrections:
            #if _debug: print '*+*+*+', wrong, correct ###DEBUG###
            if wrong == correct:
                raise SyntaxError(
                    'Detected pair of same wrong/correct tokens: %s/%s.' % ( wrong, correct, ) )
            while True:
                try:
                    indexOfWrongToken = tokens.index( wrong )
                except ValueError:
                    break
                if DEBUG: print '### %s => %s' % ( wrong, correct, )
                tokens[ indexOfWrongToken ] = correct
        return tokens

    #=======================================================================================================
    def stem( self, tokens ):
        return tokens

############################################################################################################
#
#   CAMEL CASE TOKENIZER
#
#===========================================================================================================
class CamelCaseTokenizer( Tokenizer ):

    #=======================================================================================================
    tokenCorrections = Tokenizer.tokenCorrections[:]
    tokenCorrections.append( ( 'long', 'longitude' ) )
    #-------------------------------------------------------------------------------------------------------
    uppersWithFollowingWord_re                      = re.compile( r'([A-Z]{2,})(?=[A-Z][a-z])' )
    uppersWithFollowingWord_subst                   = lambda mo: mo.group().capitalize()
    uppersOther_re                                  = re.compile( r'([A-Z]{2,})(?![a-z])' )
    uppersOther_subst                               = lambda mo: mo.group().capitalize()
    CamelCase_re                                    = re.compile( r'(?<=[a-zA-Z])(.)(?=[A-Z])' )
    #-------------------------------------------------------------------------------------------------------
    uppersWithFollowingWord_subst                   = staticmethod( uppersWithFollowingWord_subst )
    uppersOther_subst                               = staticmethod( uppersOther_subst )

    ########################################################################################################
    #   METHODS ACTING ON TEXT
    #=======================================================================================================
    def splitCamelCase( this, text ):
        R = text
        R = this.uppersWithFollowingWord_re.sub( this.uppersWithFollowingWord_subst, R )
        R = this.uppersOther_re            .sub( this.uppersOther_subst, R )
        R = this.CamelCase_re.sub( r'\1 ', R )
        return R
    splitCamelCase = classmethod( splitCamelCase )

    ########################################################################################################
    #   MAIN METHOD
    #=======================================================================================================
    def tokenize( self, text ):#, _debug = False ):
        #---------------------------------------------------------------------------------------------------
        text = self.normalizeAccentedCharacters( text )
        text = self.doPrepurgativeTextCorrections( text )
        text = self.purgeIllegalCharacters( text )
        text = self.doPostpurgativeTextCorrections( text )
        text = self.splitCamelCase( text )
        text = self.normalizeCase( text )
        #---------------------------------------------------------------------------------------------------
        tokens = self.segment( text )
        #---------------------------------------------------------------------------------------------------
        tokens = self.doTokenCorrections( tokens ) #, _debug )
        tokens = self.removeShortWords( tokens )
        tokens = self.removeStopwords( tokens )
        tokens = self.stem( tokens )
        #---------------------------------------------------------------------------------------------------
        return tokens

#===========================================================================================================
class PureCamelCaseTokenizer( CamelCaseTokenizer ):

    ########################################################################################################
    #   MAIN METHOD
    #=======================================================================================================
    def tokenize( self, text ):#, _debug = False ):
        #---------------------------------------------------------------------------------------------------
        text = self.normalizeAccentedCharacters( text )
        text = self.purgeIllegalCharacters( text )
        text = self.splitCamelCase( text )
        text = self.normalizeCase( text )
        #---------------------------------------------------------------------------------------------------
        tokens = self.segment( text )
        #---------------------------------------------------------------------------------------------------
        return tokens

############################################################################################################
#
#   PHRASE CLASSES
#
#===========================================================================================================
class PhraseABC( object ):

    #=======================================================================================================
    def distinctTokens():
        doc = """A ``Set`` of all distinct tokens in the object."""
        def fget( self ):
            return Set( self )
        return locals()
    distinctTokens = property( **distinctTokens() )

    #=======================================================================================================
    def distinctTokensByAlphabet():
        doc = """An alphabetically sorted list of all distinct tokens in the object."""
        def fget( self ):
            R = list( self.distinctTokens )
            R.sort()
            return R
        return locals()
    distinctTokensByAlphabet = property( **distinctTokensByAlphabet() )

#===========================================================================================================
class ListPhrase( PhraseABC, list ):
    pass

#===========================================================================================================
class SetPhrase( PhraseABC, Set ):
    pass

############################################################################################################
#
#
#
#===========================================================================================================
if __name__ == '__main__':

    t0 = """Kuala Lumpur, Oct. 26 (Bloomberg) – bank Negara will change the way it calculates the
    cost of lending money to commercial banks and financial institutions, the central bank said in a
    release. In a statement, the bank said that as of Nov. 1, the base lending rate, at which commercial
    banks can borrow from the central bank,will become more responsive to movements in money-market rates.
    Several weeks ago, the bank said it would change the way it calculates the base rate, said Desmond
    Ch’ng, a banking analyst OCBC Securities (Malacca) Sdn. Bhd."""

    tokens = Tokenizer().tokenize( t0 )
    phrase = ListPhrase( tokens )
    print phrase
    print phrase.distinctTokens
    print phrase.distinctTokensByAlphabet

    groveFed = u.feds.resolve( 'Grove' )
    dsGrove = groveFed.dsGrove
    tokenizer = CamelCaseTokenizer()
    view = dsGrove.catchTable( node = dict( gauge = 'abcd12' ) )
    i = -1
    for nodefact in view:
        i += 1
        if i > 50: break
        locator = nodefact.locator
        print ListPhrase( tokenizer.tokenize( locator ) ).distinctTokensByAlphabet
    text = """/CamelCase/CMFLister/ABCDIsASchema/isAGoodThing/KLM/WMSURL/4Picture/43picture/image/gif"""
    result = tokenizer.tokenize( text )
    print result
    print ListPhrase( result ).distinctTokensByAlphabet















