#
# -*- coding: utf-8 -*-

'''
$RCSfile: datasources.py,v $
$Revision: 177 $
$Author: markus $
$Date: 2005-08-22 12:50:54 +0200 (Mon, 22 Aug 2005) $
General filesystem methods all webapps need.
'''

############################################################################################################
#
#   IMPORTS
#
#===========================================================================================================

import os, string, shutil
from types import *
import biocase.configuration
from biocase.cmfinfo            import CMFListClass
from biocase.wrapper.psf        import ProviderSetup


############################################################################################################
#
#   CLASSES
#
#===========================================================================================================
 
class Datasource:
    # ------------------------------------------------------------------------------------
    def __init__(self, name, cfg=None, debug=None):
        if cfg is None:
            cfg = biocase.configuration.Cfg('configtool')
        if debug is None:
            debug = biocase.tools.debugging.DebugClass()
        self.cfg      = cfg
        self.debug    = debug
        self.name     = name
        self.psfObj   = None
        self.dbmodObj = None
        # temp references only created when needed at all
        self._slObj   = None
        
    # ------------------------------------------------------------------------------------
    def __repr__(self):
        return "DSAobj:%s"%self.name
        
    # ------------------------------------------------------------------------------------
    def getDsaDir(self):
        return os.path.join(self.cfg.datasourcesLocator, self.name)

    # ------------------------------------------------------------------------------------
    def getSchemaListObj(self):
        if self._slObj is None:
            self._slObj = CMFListClass(self.cfg, self.debug)
            self._slObj.addSchemasFromDir( self.getDsaDir() )
        return self._slObj

    # ------------------------------------------------------------------------------------
    def getSchemaNameList(self):
        slObj = self.getSchemaListObj()
        return [sObj.name for sObj in slObj]

    # ------------------------------------------------------------------------------------
    def getAbsPSFilePath(self):
        return os.path.join(self.cfg.datasourcesLocator, self.name, self.cfg.providersetupfile)

    # ------------------------------------------------------------------------------------
    def getAbsTempPSFilePath(self):
        return os.path.join(self.cfg.datasourcesLocator, self.name, self.cfg.providersetupfile[:-4]+'_tmpCfgTool.xml')

    # ------------------------------------------------------------------------------------
    def getPSFObj(self, tmp=False):
        '''Reads the PSF for a given datasource. If tmp is true read from the temporary PSF.'''
        if self.psfObj is None:
            if tmp:
                # read temporary PSF
                absFilePath = self.getAbsTempPSFilePath()
            else:
                # read real PSF
                absFilePath = self.getAbsPSFilePath()
            self.psfObj = ProviderSetup(logObj=self.debug)
            self.psfObj.setConfigDirs(PSFpath=absFilePath)
            self.psfObj.readPSFile()
        return self.psfObj

    # ------------------------------------------------------------------------------------
    def getBioCASeAccessPoint(self):
        return '%s?dsa=%s'%(self.cfg.PyWrapperURL, self.name)

    # ------------------------------------------------------------------------------------
    def getSPICEAccessPoint(self):
        return '%s?dsa=%s'%(self.cfg.PySPICEURL, self.name)

    # ------------------------------------------------------------------------------------
    def getDbmodObj(self):
        if self.dbmodObj is None:
            from biocase.wrapper.operations import getOperationsObjectForDBMod
            try:
                self.dbmodObj = getOperationsObjectForDBMod(psfObj=self.getPSFObj(), logObj=self.debug)
            except:
                self.debug + "Could not get operation module"
                self.dbmodObj = None
        return self.dbmodObj
        
    # ------------------------------------------------------------------------------------
    def getDBConnectionStatus(self):
        if self.getDbmodObj() is not None:
            return self.cfg.configtool.dbConnectionStatusGood
        else:
            return self.cfg.configtool.dbConnectionStatusBad


############################################################################################################
#
#   FUNCTIONS
#
#===========================================================================================================

# ------------------------------------------------------------------------------------
def getDsaList(cfg=None):
    if cfg is None:
        cfg = biocase.configuration.Cfg()
    dsalist = []
    for dsa in os.listdir(cfg.datasourcesLocator):
        if os.path.isdir(os.path.join(cfg.datasourcesLocator, dsa)) and os.path.isfile(os.path.join(cfg.datasourcesLocator, dsa, 'provider_setup_file.xml')):
            dsalist.append( Datasource(dsa) )
    return dsalist

# ------------------------------------------------------------------------------------
def delDsa(dsa):
    pass

# ------------------------------------------------------------------------------------
def createDsa(dsa, cfg=None, debug=None):
    '''Create a new datasource directory and copy the empty skeleton files to it.'''
    if cfg is None:
        cfg = biocase.configuration.Cfg()
    if debug is None:
        debug = biocase.tools.debugging.DebugClass()
    # check dsa spelling. replace spaces with underscore
    dsa = dsa.replace(' ', '_')
    # abs path to new dir
    dsaDir = os.path.join(cfg.datasourcesLocator, dsa)
    # does it already exist?
    if os.path.isdir( dsaDir ):
        debug + "Datasource %s does already exist!"%dsa
        return False
    else:
        # create directory and copy files recursively
        shutil.copytree( os.path.join(cfg.configurationLocator, 'configtool','dsa_skeleton'), dsaDir )
        debug + "Datasource %s created successfully!"%dsa
        return True
    return False
    
    