#
# -*- coding: utf-8 -*-

''' TBD '''


############################################################################################################
#
#   IMPORTS
#
#===========================================================================================================
from biocase.querytool import __version__
from biocase.tools.htmltools import escapeHtml
from biocase.wrapper.operations import recordStatus
from biocase.querytool.querygenerator import QueryGenerator
from mx.DateTime import now
import urllib
import logging
# debugging to file
log = logging.getLogger("webapp.querytool")

import libxml2
# xmlDOM.name : returns the node name
#      * type : returns a string indicating the node type
#      * content : returns the content of the node, it is based on xmlNodeGetContent() and hence is recursive.
#      * parent , children, last, next, prev, doc, properties: pointing to the associated element in the tree, those may return None in case no such link exists.




def namespaceDefs(node):
    n = node.nsDefs()
    while n:
        yield n
        n = n.next

############################################################################################################
#
#   ERROR CLASSES
#
#===========================================================================================================
class WrapperError(Exception):
    def __init__(self):
        pass
        

############################################################################################################
#
#   CLASSES
#
#===========================================================================================================
class QueryDispatcher:
    '''A class to send queries to wrappers and do simple work with protocol responses.
    Supports DiGIR and BioCASe'''
    
    # ------------------------------------------------------------------------------------
    def __init__(self, protocolNS=None):
        if protocolNS is None:
            protocolNS = QueryGenerator.NS_BioCASe
        elif protocolNS.lower() == 'biocase':
            protocolNS = QueryGenerator.NS_BioCASe
        elif protocolNS.lower() == 'digir':
            protocolNS = QueryGenerator.NS_DiGIR            
        elif protocolNS.lower() == 'tapir':
            protocolNS = QueryGenerator.NS_TAPIR
        self._type          = protocolNS
        self.query          = None # the last query XML protocol string used
        self.response       = None # the response XML protocol string
        self.doc            = None # the response as a DOM object
        self.contentDOM     = None # the content DOM element of the response
        self.diagnosticsDOM = None # the diagnostics DOM element of the response
        self.recStatus      = recordStatus()
        
    # ------------------------------------------------------------------------------------
    def sendQuery(self, wrapper_url, protocolString, debuglevel=1, security_role=None):
        '''Sends a query to the wrapper and returns a record status object or None if query failed.'''
        self.query = protocolString
        # if digir destination parse resource of form RESOURCE@ACCESSPOINT
        if wrapper_url.count("@") == 1:
            digirResource, wrapper_url = wrapper_url.split("@")
            log.info("DiGIR resource %s, embedded in wrapper url, cut off." % (digirResource))
        log.info("SEND QUERY TO %s"%wrapper_url)
        log.info("SECURITY ROLE=%s"%str(security_role))
        
        #We query with the result XML to the wrapper:
        # @request is used by DiGIR, @query by biocase
        if self._type == QueryGenerator.NS_DiGIR:
            params = {'request': protocolString.encode('utf-8')}
        else:
            params = {'query': protocolString.encode('utf-8'), 'debug':debuglevel}
        if security_role is not None:
            params['role'] = security_role
        params = urllib.urlencode(params)
        #
        # risky. use error handling
        try:
            f = urllib.urlopen(wrapper_url, params)
            self.response = f.read()
            log.info("FULL WRAPPER RESPONSE %s"%unicode(self.response, 'utf-8'))
        except:
            log.info("HTTP ERROR: WRAPPER URL %s COULD NOT BE OPENED" % wrapper_url)
            return None
        # parse the xml into xmllib2 DOM:
        try:
            self.doc = libxml2.parseDoc(self.response)
        except:
            self.doc = None
            log.info("ERROR: NO WELL FORMED XML WRAPPER RESPONSE FOUND")
            return None
        log.info("ANALYZING RESPONSE DOC")
        root = self.doc.getRootElement()
        log.info("  DOC ROOT %s"%root.name)
        if root.ns() is not None:
            try:
                self._type = unicode(root.ns()).replace('"','').replace("'",'')
                self._type = self._type[self._type.find('=')+1:]
                log.info("  PROTOCOL NS found: %s" % self._type)
            except:
                self._type = None
        else:
            log.info("  PROTOCOL NS NOT found. ROOT ATTRIBUTES: %s" % root.get_properties())
        for header in root.children:
            if header.type == 'element' and header.name=='header':
                log.info("  Header section found.")
                self.headerDOM = header
                break
        # try to get main sections
        elem = header
        while elem is not None:
            if elem.type == "element":
                if elem.name == 'content':
                    log.info("  Content section found.")
                    self.contentDOM = elem
                # try to get diagnostics
                elif elem.name == 'diagnostics':
                    log.info("  Diagnostics section found.")
                    self.diagnosticsDOM = elem
            elem = elem.next

        # try to get record status (paging info)

        # DiGIR keeps record status in the diagnostics
        if self._type == QueryGenerator.NS_DiGIR:
            for severity, code, diag in self.getDiagnostics():
                if code == 'MATCH_COUNT':
                    self.recStatus.total = int( diag )
                    log.info("    Record Total = %i"%(self.recStatus.total))
                elif code == 'RECORD_COUNT':
                    self.recStatus.count = int( diag )
                    log.info("    Record Count = %i"%(self.recStatus.count))
                elif code == 'END_OF_RECORDS':
                    pass
        # BioCASE
        elif self._type == QueryGenerator.NS_BioCASe and self.contentDOM is not None:
            self.recStatus.start = int( self.contentDOM.prop('recordStart') )
            log.info("    Record Start = %i"%(self.recStatus.start))
            self.recStatus.count = int( self.contentDOM.prop('recordCount') )
            log.info("    Record Count = %i"%(self.recStatus.count))
            self.recStatus.drop = int( self.contentDOM.prop('recordDropped') )
            log.info("    Record Drop = %i"%(self.recStatus.drop))
            try:
                self.recStatus.total = int( self.contentDOM.prop('totalSearchHits') )
                log.info("    Record Total = %i"%(self.recStatus.total))
            except:
                self.recStatus.total = None                        
                log.info("    Record Total is not existing")
        return self.recStatus
        
    # ------------------------------------------------------------------------------------
    def getScanValues(self):
        if self.contentDOM is None or self.contentDOM.children is None:
            return []
        result = []
        for val in self.contentDOM.children:
            # biocase
            if val.type == "element" and val.name == 'value':
                result.append( unicode(val.content, 'UTF-8' ) )
            # digir
            elif val.type == "element" and val.name == 'record':
                for n in val:
                    if n.type == "element":
                        result.append( unicode(n.content, 'UTF-8' ) )
                        break
        return result

    # ------------------------------------------------------------------------------------
    def getContent(self, asDoc=True):
        if self.contentDOM is None or self.contentDOM.children is None:
            log.info("COULDN'T GET CONTENT DOM NODE.")
            return None
        contentDOM = self.contentDOM.copyNodeList()
        n = contentDOM.children
        while n is not None:
            nextnode = n.next
            if n.type == 'element':
                log.info("  Content payload element %s found." % n.name)
            else:
                log.info("  Non payload element %s (type %s) ignored."%(n.name, n.type))
                n.freeNode()
                n.unlinkNode()
            n = nextnode
        # check how many payload elements content has...
        log.info("  ContentDOM.lsCountNode() = %i" % contentDOM.lsCountNode())
        if contentDOM.lsCountNode() < 1:
            log.info("NO CONTENT FOUND.")
            return None
        
        if contentDOM.lsCountNode() == 1:
            payload = contentDOM.children
        else:
            # more than one child element node of content. create artificial recordset node for this digir style
            payload = contentDOM
            payload.setName('RecordSet')
            payload.setNs( contentDOM.children.ns() )
        
        # add darwin core NS for digir
        if self._type == QueryGenerator.NS_DiGIR:
            payload.setProp('xmlns:darwin','http://digir.net/schema/conceptual/darwin/2003/1.0')
            payload.setProp('xmlns:xsd','http://www.w3.org/2001/XMLSchema')
            payload.setProp('xmlns:xsi','http://www.w3.org/2001/XMLSchema-instance')

        # create new doc with these nodes:
        if asDoc:
            doc = libxml2.newDoc("1.0")
            doc.setRootElement(payload)
        else:
            doc = payload

        log.info("CONTENT FOUND: %s" % unicode(doc))
        return doc

    # ------------------------------------------------------------------------------------
    def getDiagnostics(self):
        '''Return a list of triplet tuples of severity,code,content.'''
        diagnostics = []
        if self.diagnosticsDOM is not None and self.diagnosticsDOM.children is not None:
            for diag in self.diagnosticsDOM.children:
                if diag.type == "element" and diag.name == 'diagnostic':
                    if diag.hasProp('severity'):
                        severity = diag.prop('severity')
                    else:
                        severity = None
                    if diag.hasProp('code'):
                        code = diag.prop('code')
                    else:
                        code = None
                    diagnostics.append( (severity, code, unicode(diag.content, 'UTF-8' )) )
        return diagnostics



    
