# -*- coding: utf8 -*-
'''
LOGGING MODULE
********************
***  LOGGING MODULE
********************
$RCSfile: logging.py,v $
$Revision: 569 $
$Author: markus $
$Date: 2006-03-09 16:10:04 +0100 (Thu, 09 Mar 2006) $

Module containing a simple logging class for errors and activity.
Refer to the class variables loggingStates and loggingRates for available options.
'''


from mx.DateTime import now, gmt
import traceback
import sys
from os.path 	import join
from string 	import join as strjoin, upper
from types		import *
import biocase.wrapper


class logEntry:
	'''A single log entry for the logging class'''
	# the diagnostics/logging type starting with 0=error and detailed debugging=4
	LOG_TYPES      = ['error','warn','info','debug','debug']
	def __init__(self, dataDict):
		self.timestamp = now()
		self.sysinfo   = None
		self.log_type  = 0
		self.errorcode = None
		self.message   = None
	
	def __str__(self):
		'''Return the entry in a simple formatted way.'''
		errCode = ''
		if self.errorcode is not None:
			errCode = "%s: "%str(self.errorcode)
		logstring = "%s  %s  %s\n" %(self.timestamp, errCode, self.message)
		logstring += unicode(self.sysinfo, errors='replace')
		return logstring
		
	
class loggingClass:
	'''This is a simple logging class to keep track of occured errors.
	Add log messages by uses the + operator. Timestamps and system errormessages will automatically be added.
	You can add a tuple (debuglevel<integer>, logmessage<string>) to the logging instance indicating the importance of the logging message.
	The higher the debuglevel, the less important the message is.
	When adding only a message as a string, the highest debuglevel(=0) is assigned by default to the message.
	
	self.debuglevel is the level used for XML output.
	
	The logging parameters .lograte and .logstate influence the file.logging activity.
	'''
	LOG_TYPES      = ['debug','info','warn','error','fatal']
	LOG_LEVEL      = {0:'critical', 1:'performance', 2:'simple debug', 3:'detailed debug'}
	loggingStates  = {'off':0, 'simple':1, 'detailed':2}
	loggingRates   = {'daily':0, 'weekly':1, 'monthly':2}
	logLevelLimit  = 100

	def __init__(self):
		# logging list. contains tuples: 1-gmtime, 2-localtime, 3-log entry, 4-traceback error
		self.log = [] 
		self.lastTimeLog   = now()
		# stores last traceback so it will be displayed only once
		self.lastTraceback = 0
		# max severity level of loggings (errors). 0=no error, 1=warning, 2=error handled, 3=severe error
		self.error = 0
		# settings
		self.logdir	 = None
		self.logfileBasename = 'requests'
		self.debuglevel = 0
		self.lograte	= 1
		self.logstate   = 1
		# wrapper specific stuff for error logging
		self.sourcename   = 'unknown' # the subfolder of database we are in
		self.responseCount = '?'
		self.responseDrop  = '?'
		self.responseTotal = '?'
		self.filterString = 'No request object could be established'
		self.requestType  = 'unknown'
		self.requestStart = '?'
		self.requestLimit = '?'
		self.requestDoc   = 'no request doc set.'
		self.sources	  = [] # original sources of the request
		# log logging start
		self+(1,"Logging started.")


	def __add__(self, log):
		'''operator overload to add a logging message to the instance.
		Accepts a tuple (debuglevel<integer>, logmessage<string>)
		or a plain string as parameter 'log'.
		'''
		if isinstance(log, loggingClass):
			# add other log class
			for entry in log.log[1:]:
				self.log.append(entry)
		else:
			debuglevel  = 0
			log_type    = 0
			noerror	    = 0
			errorcode   = None
			log_message = None
			error       = None
			errorMsg    = ""
			if type(log) in (StringType, UnicodeType):
				log_message = log
			elif type(log) == TupleType:
				# old style of passing parameters
				if len(log) > 2:
					noerror = log[2]
				debuglevel = log[0]
				log_message = log[1]
			elif type(log) ==DictType:
				# new parameter type with error coders, etc. should be used from now on
				if log.has_key('type'):
					log_type = log['type']
				if log.has_key('level'):
					debuglevel = log['level']
				if log.has_key('code'):
					errorcode= log['code']
				if log.has_key('msg'):
					log_message = log['msg']
				if log.has_key('noerror'):
					noerror = log['noerror']
			# check traceback info
			error = sys.exc_info()
			if error[2] != None and self.lastTraceback != error:
				self.lastTraceback = error
				# is error traceback not suppressed?
				if noerror == 0:
					# log error
					log_type = 3
					self.error = 2
					errorMsg = "".join( traceback.format_exception( error[0], error[1], error[2] ) )
			if type(log_message) <> UnicodeType:
				log_message = unicode(log_message, errors='replace')
			logInfo = (gmt(), now(), log_message, errorMsg, debuglevel, errorcode, log_type, error)
			self.log.append(logInfo)
	
	def __repr__(self):
		'''returns logs according to the debugging level as a string'''
		logging = u'The Logging Results:\n'
		self.log.sort(compareLogClassEntries)
		for entry in self.log:
			if entry[4] <= self.debuglevel:
				logging += ('-' * 22) + '\n'
				if entry[5] is not None:
					errCode = "%s: "%str(entry[5])
				else:
					errCode = ''
				logging += "%s  %s  %s\n" %(entry[0], errCode, entry[2])
				logging += unicode(entry[3], errors='replace')
		return logging
	
	def setDebugLevel(self, debuglevel):
		self + (2, "Setting debuglevel from %i to %i"%(self.debuglevel, debuglevel) )
		self.debuglevel = debuglevel
		
	def setRequestData(self, requestObj):
		'''Takes a request object and retrieves relevant data from it for logging'''
		try:
			self.filterString	= unicode(requestObj.filter)
			self.requestType	= requestObj.type
			if requestObj.count > 0:
				self.requestType = "count"
			if self.requestType == "scan":
				self.filterString = "Scan: %s with filter: %s" %(requestObj.scan, unicode(requestObj.filter))
			self.requestStart   = requestObj.start
			self.requestLimit   = requestObj.limit
			self.sources		= strjoin(requestObj.sources, ' <- ')
		except:
			self + "Could not log request data"
		
	def setSourcename(self, name):
		'''Sets the name of the source being queried.'''
		if len(name)>0:
			self.sourcename = name

	def setResponseStatus(self, count='?', drop='?', total='?'):
		'''set the response status data of the logging. Use count for count requests.'''
		self.responseCount = count
		self.responseDrop  = drop
		self.responseTotal = total

	def setError(self, level):
		'''Sets the max error level self.error of the logging.'''
		self.error = max(self.error, level)
		
	def returnLogList(self):
		'''returns logs according to the debugging level 
		as a list of tuples (error_code, log_message)'''
		result=[]
		self.log.sort(compareLogClassEntries)
		for entry in self.log:
			if entry[4] <= self.debuglevel:
				result.append( (self.LOG_TYPES[entry[6]], entry[5], entry[2] ) )
				(tbo1, tbo2, tbo3) = entry[7]
				if tbo1 is not None and self.debuglevel > 0:
					# append last traceback line only
					result.append( (self.LOG_TYPES[entry[6]], entry[5], strjoin( traceback.format_exception_only(tbo1, tbo2) ) ) )
					# append the whole traceback
					tracebackText = unicode(entry[3], errors='replace')
					result.append( (self.LOG_TYPES[0], entry[5], tracebackText ) )
					pass
			#else:
			#	result.append( (self.LOG_TYPES[entry[6]], entry[5], "DROPPED: %s" % entry[2] ) )
		return result
		
	def logPassedTime(self, title='unknown process'):
		'''log the time it took to process a task from the last timelog and the full start to now.'''
		if self.debuglevel < 0:
			timeNow   = now()
			sinceLast = timeNow - self.lastTimeLog
			totalTime = timeNow - self.log[0][1]
			self.lastTimeLog = timeNow
			self + (-1 , "Processing '%s' took %s. Total processing time: %s." %(title, str(sinceLast), str(totalTime)))

	def logErrors(self):
		# try to open the error-logfile for writing
		if self.logdir == None:
			self + "No log directory specified - no error logfile will be created."
		else:
			try:
				logname  = self.getLogFileName('error')+'.log'
				filename = join(self.logdir, logname)
				if self.error > 1:
					# open file
					logfile = file(filename, 'a')
					# error occured. write log.
					logfile.write("\n%s %s %s" %("*"*40, upper(self.sourcename), "*"*40))
					logfile.write("\n### ERROR TIME %s"%(str(self.log[0][1])))
					logfile.write("\n### WRAPPER VERSION %s"%(biocase.wrapper.__version__))
					logfile.write("\n### REQUEST from %s to %s" %(self.sources, self.sourcename ))
					logfile.write("\n" + self.requestDoc)		
					logfile.write("\n### ERROR LOG\n")
					# give a detailed logging. temporarily change debuglevel:
					oldlevel = self.debuglevel
					self.setDebugLevel( 3 )
					logfile.write(str(self))
					self.setDebugLevel( oldlevel )
					# footer
					logfile.write("\n\n")
					logfile.close()
			except:
				self + "No write permission for error-log file exists. No error logging done."
	
	def writeLogFile(self):
		'''Writes an error logfile if an error occured and a log file if logging activity (state) is set > 0.
		'''
		# try to open the error-logfile for writing
		self.logErrors()
		if self.logstate == 0:
			# no logging
			self + "Logging deactivated."
		elif self.logdir == None:
			self + "No log directory specified - no logfile will be created."
		else:
			# logging active, try to open the logfile for writing
			try:
				# build filename
				logname  = self.getLogFileName(self.logfileBasename)+'.log'
				filename = join(self.logdir, logname)
				# open file
				logfile = file(filename, 'a')
				# get return status
				if self.error == 0:
					status = 'OK'
				elif self.error == 1:
					status = 'WARNING'
				else:
					status = 'ERROR'
				# simple 1-liner
				if self.requestType == "scan":
					headline = "%s: %s - %s request from %s. Process time: %s. Return status: %s. %s distinct records.  Scanned concept: %s." %( upper(self.sourcename), str(self.log[0][1]), self.requestType, self.sources, str(self.getProcessingTime()), status, self.responseCount, self.filterString )
				elif self.requestType == "search":
					headline = "%s: %s - %s request with start=%s, limit=%s from %s. Process time: %s. Return status: %s. %s/%s/%s returned/dropped/matched records.  Filter: %s." %( upper(self.sourcename), str(self.log[0][1]), self.requestType, self.requestStart, self.requestLimit, self.sources, str(self.getProcessingTime()), status, self.responseCount, self.responseDrop, self.responseTotal, self.filterString )
				else:
					headline = "%s: %s - %s request from %s. Process time: %s. Return status: %s. %s/%s/%s returned/dropped/matched records.  Filter: %s." %( upper(self.sourcename), str(self.log[0][1]), self.requestType, self.sources, str(self.getProcessingTime()), status, self.responseCount, self.responseDrop, self.responseTotal, self.filterString )
				headline = headline.encode('utf-8')
				if self.logstate == 1:
					# simple logging.
					logfile.write("\n"+headline)
				else:
					logfile.write("\n"+"*"*80)
					logfile.write("\n%s\n" %headline)
					logfile.write(self.requestDoc)		
					logfile.write("\n")		
				logfile.close()
			except:
				self + "No write permission for log file exists. No file logging done."
				raise

	def writeFullLogFile(self, basename=None):
		if self.logdir == None:
			self + "No log directory specified - no logfile will be created."
		else:
			# build filename
			if basename is None:
					basename=self.logfileBasename
			logname  = self.getLogFileName(basename)+'.log'
			filename = join(self.logdir, logname)
			# open file
			logfile = file(filename, 'a')
			logfile.write("\n\n\n")
			logfile.write(self.__repr__())
			logfile.close()
			
	def getProcessingTime(self):
		'''Return the duration from the beginning of the logging to now.'''
		return now()-self.log[0][1]
		
	def getLogFileName(self, basename='wrapper'):
		filename = basename
		if self.lograte == 0:
			# daily
			filename += now().strftime("-%Y-%b-%d")
		elif self.lograte == 1:
			# weekly
			filename += now().strftime("-%Y-week%U")
		elif self.lograte == 2:
			# monthly
			filename += now().strftime("-%Y-%b")
		else:
			# yearly
			filename += now().strftime("-%Y")
		return filename
	
	def getLograte(self):
		getRate={}
		for real, val in self.__class__.loggingRates.items():
			getRate[val]= real
		return getRate[self.lograte]
	
	def getLogstate(self):
		getState={}
		for real, val in self.__class__.loggingStates.items():
			getState[val]= real
		return getState[self.logstate]
	

def compareLogClassEntries(arg1, arg2):
	if arg1[0] > arg2[0]:
		return 1
	elif arg1[0] < arg2[0]:
		return -1
	else:
		return 0




if __name__ ==  "__main__":
	import sys
	import traceback as t
	log = loggingClass()
	log.setDebugLevel( 2 )
	log + (1,"hi bert")
	log + (1,"i am third")

	b = loggingClass()
	b.setDebugLevel( log.debuglevel )
	b.logPassedTime("B stamp")
	b + "B:hi bert"
	b + (2,"B:i am third")
	b.logPassedTime("B stamp")
	log + (1,"kung fu fighting")
	log+b
	log.logPassedTime("LOG stamp")
	for i in range(100000):
		x=i+1
	log.logPassedTime("LOG stamp 2")

	try:
		def crash():
			return 1/0
	except:
		pass
	
	try:
		print crash()
	except:
		log + (1,"duno")

	log + (1,"another time its NOW.")

	print log

	print log.returnLogList()
