# -*- coding: utf8 -*-
'''
METADATA MODULE
**********************
***  METADATA MODULE
**********************
$RCSfile: metadata.py,v $
$Revision: 9 $
$Author: markus $
$Date: 2005-07-04 11:47:00 +0200 (Mon, 04 Jul 2005) $

Functions associated in reading, writing and parsing 
the datasource metadata.
'''

import os
import os.path
import xml.dom.ext
from xml.sax	import parse
from biocase.wrapper.errorclasses 		import *
from biocase.wrapper.logging 			import loggingClass
from biocase.tools.various_functions 	import unique, isTrue, determineServerIP
from biocase.wrapper.metadata_handler 	import DsaMetadataConfigHandler
from biocase.cmfinfo                    import CMFListClass, CMFInfoClass

		

# ----------------------------------------------------------------------------
class contact:
	'''A single related entity contact.'''
	def __init__(self, logObj=loggingClass()):
		# store reference to logging object
		self.logObj = logObj
		#
		# data
		#
		self.type        = None
		self.name        = None
		self.title       = None
		self.email       = None
		self.phone       = None



# ----------------------------------------------------------------------------
class relEntities:
	'''A single related entity of a datasource.'''
	def __init__(self, logObj=loggingClass()):
		# store reference to logging object
		self.logObj = logObj
		#
		# data
		#
		self.ID          = None
		self.name        = None
		self.acronym     = None
		self.logoURL     = None
		self.role        = [] # strings
		self.description = None
		self.relInfoURLs = [] # strings
		self.contacts    = [] # contact objects


# ----------------------------------------------------------------------------
class DsaMetadata:
	namespace = ["http://www.biocase.org/schemas/dsa_metadata/1.0"]

	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def __init__(self, cfg=None, debug=None):
		# store reference to logging object
		if cfg is None:
			cfg = biocase.configuration.Cfg()
		if debug is None:
			debug = biocase.tools.debugging.DebugClass()
		self.cfg = cfg
		self.debug=debug
		# the absolute filepath to the metadata file
		self.file		 = None
		#
		# data
		#
		self.name        = None
		self.label       = None
		self.accessPoint = None
		self.abstract    = None
		self.keywords    = None
		self.citation    = None
		self.rights      = None
		# related entities
		self.CMFlist     = CMFListClass()
		self.relatedEntities = {} # key=entity name, value = relEntity object

	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def repr(self):
		from biocase.tools.showobject import view
		return view(self)	
	
	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def addRelEntity(self, relEntObj):
		'''Sets the datasource name and reads the existing schemas.'''
		self.relatedEntities[relEntObj.name] = relEntObj
				
	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def setDsaDir(self, dsa):
		'''Sets the datasource name and reads the existing schemas.'''
		if dsa is not None:
			self.CMFlist.addSchemasFromDir( os.path.join(self.cfg.datasourcesLocator, dsa) )
				
	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def readFile(self, absFilename=None):
		#
		# use SAX to parse the wrapperconfig.xml file and fill this object
		if absFilename == None:
			absFilename = self.file
		try:
			handler = DsaMetadataConfigHandler(self, self.debug, self)
			parse(absFilename, handler)
		except:
			raise "read metadata error"


	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def writeFile(self, filename=None, namespace=None):
		'''Writes the PSF object to an xml file.
		If no namespace is given it uses by default the latest PSF version.'''
		if filename is None:
			filename = self.file
		MetadataFile = file(filename,'w')
		self.debug + "writing PSFile to %s" %(filename)
		xml.dom.ext.PrettyPrint(self.createDOM(version=namespace), stream=MetadataFile, encoding='utf-8')
		MetadataFile.close()

	# ------------------------------------	
	# write methods
	# ------------------------------------	

	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def createDOM(self, version=None):
		'''	Returns a minidom object that represents this metadat object as XML.'''
		from biocase.tools.xmlwriter import XMLwriterClass
		if version == None:
			# by default use the latest version
			version = self.__class__.namespace[-1]
		if version == self.__class__.namespace[0]:
			self.XMLwriter = XMLwriterClass(self.__class__.namespace[0])
			dom = self.createDOM_1_0()
		else:
			raise "unknown version. Failed to write XML."
		# free memory
		self.XMLwriter = None
		return dom
		
	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def createDOM_1_0(self):
		'''Return a dsa metadata dom object with namespace version 1.0'''
		root_node = self.XMLwriter.createNode("Metadata")
		# HOST
		host_node = self.XMLwriter.createNode("Host")
		root_node.appendChild(host_node)
		self.XMLwriter.addNode(host_node, "User", self.user)
		self.XMLwriter.addNode(host_node, "Password", self.password)
		self.XMLwriter.addNode(host_node, "Datasource", self.datasource)
		self.XMLwriter.addNode(host_node, "DB_IP", self.dbIP)
		self.XMLwriter.addNode(host_node, "DBMS", self.dbms)
		# DB
		db_node = self.XMLwriter.createNode("DB")
		root_node.appendChild(db_node)
		self.XMLwriter.addNode(db_node, "TextEncoding", self.encoding)
		# table aliases
		tables = self.getTableAliasList()
		for t in tables:
			ta_node = self.XMLwriter.createNode("TableAlias", None, {'tablename':t.table, 'alias':t.alias})
			db_node.appendChild(ta_node)
			# PKS
			pk_node = self.XMLwriter.createNode("PrimaryKey")
			ta_node.appendChild(pk_node)
			for pk in t.pk:
				self.XMLwriter.addNode(pk_node, 'Attribute', pk.attr, {'dbType':pk.type})
			# FKS
			for fkObj in t.fks:
				fk_node = self.XMLwriter.createNode("ForeignKey", None, {'target':fkObj.target})
				ta_node.appendChild(fk_node)
				for attrObj in fkObj.attrList:
					self.XMLwriter.addNode(fk_node, 'Attribute', attrObj.attr, {'dbType':attrObj.type})
		# SCHEMAS
		schemas_node = self.XMLwriter.createNode("SupportedSchemas")
		root_node.appendChild(schemas_node)
		for ss in self.schemas.values():
			ss_node = self.XMLwriter.createNode("Schema", os.path.split(ss.conceptMappingFile)[1], {'namespace':ss.name, 'request':ss.request, 'response':ss.response})
			schemas_node.appendChild(ss_node)
		# WRAPPER BEHAVIOR
		behavior_node = self.XMLwriter.createNode("WrapperBehavior")
		root_node.appendChild(behavior_node)
		
		return root_node

