# -*- coding: utf8 -*-
'''
**********************************************
*	   P R O T O C O L - C L A S S
*	   for the biocase protocol
**********************************************
$RCSfile: biocase_response.py,v $
$Revision: 218 $
$Author: markus $
$Date: 2005-08-25 18:24:18 +0200 (Thu, 25 Aug 2005) $

Request and Response classes to support the biocase protocol.
'''

import types, sys, os
import xml.dom.minidom, xml.dom.ext
from mx.DateTime 			import now
from xml.sax							import parse, parseString
from cgi								 import FieldStorage
from os								 import environ
from os.path							 import isfile

import biocase.wrapper
from biocase.tools.various_functions 		import isTrue
from biocase.wrapper.typeconverter	import convertDateTime2ISOString
from biocase.wrapper.cmf_base		import CMFClass
from biocase.wrapper.operations     import recordStatus
from biocase.wrapper.errorclasses		 import *
from biocase.wrapper.protocol.BioCASe.biocase_handler	 import requestHandler
from biocase.wrapper.protocol.base_request	 import RequestBaseClass
from biocase.wrapper.protocol.base_response	 import ResponseBaseClass
from biocase.wrapper.sql.operators		 import compressOperator

	


# ------------------------------------------------------------------------------------
class ResponseClass(ResponseBaseClass):
	'''Response class implementing teh biocase protocol for responses:
	http:www.biocase.org/protocol
	'''
	# protocol namespace and url:
	__protocol__	= u'BioCASE Unitlevel Access Protocol version 1.31'
	__protocolURL__ = u'http://www.bgbm.org/biodivinf/schema/protocol_1_31.xsd'
	__protocolNS__  = u'http://www.biocase.org/schemas/protocol/1.3'

	# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	def __init__(self, logObj):
		ResponseBaseClass.__init__(self, logObj)
		# status about paging, record count etc.
		self.recordStatus = recordStatus()
		
	# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	def init(self):
		'''Overriden method called by the base class to initialize the raw protocol response doc.'''
		# root element
		response_elem = self.doc.createElementNS(self.__class__.__protocolNS__,"response") # the root element
		response_elem.setAttribute('xmlns:xsi', 'http://www.w3.org/2001/XMLSchema-instance')
		response_elem.setAttribute('xsi:schemaLocation', self.__class__.__protocolNS__ +' ' +self.__class__.__protocolURL__)
		self.doc.appendChild(response_elem)
		# header information:
		header_elem = self.doc.createElementNS(self.__class__.__protocolNS__,"header") # the root element
		response_elem.appendChild(header_elem)
		tmp_elem = self.doc.createElementNS(self.__class__.__protocolNS__,"version")
		tmp_elem.setAttribute('software', 'Python Interpreter')
		header_elem.appendChild(tmp_elem)
		text = self.doc.createTextNode(sys.version)
		tmp_elem.appendChild(text)
		try:
			tmp_elem = self.doc.createElementNS(self.__class__.__protocolNS__,"version")
			tmp_elem.setAttribute('software', 'Wrapper')
			text = self.doc.createTextNode(biocase.wrapper.__version__)
			tmp_elem.appendChild(text)
			header_elem.appendChild(tmp_elem)
		except:
			self.logObj+(1,'Wrapper software version could not be determined')
		try:
			tmp_elem = self.doc.createElementNS(self.__class__.__protocolNS__,"version")
			tmp_elem.setAttribute('software', 'DB module')
			header_elem.appendChild(tmp_elem)
		except:
			pass

		try:
			tmp_elem = self.doc.createElementNS(self.__class__.__protocolNS__,"version")
			tmp_elem.setAttribute('software', 'OS')
			text = self.doc.createTextNode(os.name)
			tmp_elem.appendChild(text)
			header_elem.appendChild(tmp_elem)
		except:
			pass
		
		# create empty sendTime element to be filled later when XML is returned
		tmp_elem = self.doc.createElementNS(self.__class__.__protocolNS__,"sendTime")
		header_elem.appendChild(tmp_elem)
			
		# create empty type element to be filled later
		tmp_elem = self.doc.createElementNS(self.__class__.__protocolNS__,"type")
		header_elem.appendChild(tmp_elem)
		
		# create diagnostics root element
		ELdiagnostics = self.doc.createElementNS(self.__class__.__protocolNS__,"diagnostics")
		response_elem.appendChild(ELdiagnostics)

	# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	def setPsfObj(self, psfObj):
		'''Adds a psf object to the protocol object and updates 
		the initial protocol DOM doc with the PSF information.'''
		# store reference to psf object through base method
		ResponseBaseClass.setPsfObj(self, psfObj)
		#
		# biocase specific ones
		#
		# get header element
		header_elem = self.doc.documentElement.firstChild
		type_elem = header_elem.getElementsByTagName("type")[0]
		# add source element
		try:
			tmp_elem = self.doc.createElementNS(self.__class__.__protocolNS__,"source")
			text = self.doc.createTextNode(self.psfObj.IP)
			tmp_elem.appendChild(text)
			header_elem.insertBefore(tmp_elem, type_elem)
		except:
			self.logObj+(1,'Source hostname could not be determined')

	# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	def setRequestObj(self, requestObj=None):
		'''Adds a request object to the protocol object and updates 
		the initial protocol DOM doc with the PSF information.'''
		# store reference to request object through base method
		ResponseBaseClass.setRequestObj(self, requestObj)
		#
		# biocase specific preparation of protocol doc
		#
		# get header element
		header_elem = self.doc.documentElement.firstChild
		type_elem = header_elem.getElementsByTagName("type")[0]

		# add destination element
		if len(self.requestObj.sources) > 0:
			for dest in self.requestObj.sources:
				tmp_elem = self.doc.createElementNS(self.__class__.__protocolNS__,"destination")
				text = self.doc.createTextNode(dest)
				tmp_elem.appendChild(text)
				header_elem.insertBefore(tmp_elem, type_elem)
		else:
			tmp_elem = self.doc.createElementNS(self.__class__.__protocolNS__,"destination")
			text = self.doc.createTextNode('unknown')
			tmp_elem.appendChild(text)
			self.logObj+(1,'Destination hostname could not be determined')
			header_elem.insertBefore(tmp_elem, type_elem)

	# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	def setCapabilitiesContent(self):
		self.type = "capabilities"
		# create capabilities XML doc:
		capa_elem = self.doc.createElementNS(self.__class__.__protocolNS__,"capabilities") # the root element
		for schema in self.psfObj.schemas.values():
			# add all mapped attributes of DB into Concepts elements.
			# Use the full CMF class to read CMFs:
			CMFobj = CMFClass(self.logObj)
			try:
				CMFobj.loadCMFdata(schema.conceptMappingFile, self.psfObj)
			except CouldNotPickleError:
				pass
			except:
				CMFError(cmf=schema.name, log=self.logObj)
				continue
			# create schema element
			tmp_elem = self.doc.createElementNS(self.__class__.__protocolNS__,"SupportedSchemas")
			tmp_elem.setAttribute('namespace', schema.name)
			if isTrue(schema.request):
				tmp_elem.setAttribute('request', 'true')
			else:
				tmp_elem.setAttribute('request', 'false')
			if isTrue(schema.response):
				tmp_elem.setAttribute('response', 'true')
			else:
				tmp_elem.setAttribute('response', 'false')
			# loop thru concepts and output each xpath plus datatype & searchable in diagnostics
			concepts = CMFobj.path2mappings.keys()
			concepts.sort()
			for path in concepts:
				CMFelem = CMFobj.path2ElemObj[path]
				concept_elem = self.doc.createElementNS(self.__class__.__protocolNS__,"Concept")
				# use path as element text
				text = self.doc.createTextNode(path)
				concept_elem.appendChild(text)
				# is it searchable ?
				if CMFelem.isSearchable():
					searchable = "true"
				else:
					searchable = "false"
				concept_elem.setAttribute('searchable', searchable)
				# which w3c data type does it have?
				if CMFelem.Type is not None:
					concept_elem.setAttribute('datatype', CMFelem.Type)
				# append cocept element to schema node
				tmp_elem.appendChild(concept_elem)
			capa_elem.appendChild(tmp_elem)
		self.content = capa_elem

	# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	def setCountContent(self, x):
		self.type = "count"
		# create count XML doc:
		count = self.doc.createElementNS(self.__class__.__protocolNS__,"count") # the root element
		text = self.doc.createTextNode(str(x))
		count.appendChild(text)
		# set content property
		self.content = count
		
	# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	def setScanContent(self, list):
		self.type = "scan"
		# create scan XML doc:
		scan_elem = self.doc.createElementNS(self.__class__.__protocolNS__,"scan") # the root element
		# set recordCount attribute value
		self.recordStatus.count = len(list)
		self.recordStatus.start = 0
		self.recordStatus.drop  = 0
		self.recordStatus.total = None
		# create XML scan list
		list.sort()
		for val in list:
			if val != None and len(val) > 0:
				val_elem = self.doc.createElementNS(self.__class__.__protocolNS__,"value")
				text = self.doc.createTextNode( val )
				val_elem.appendChild(text)
				scan_elem.appendChild(val_elem)
		# set content property
		self.content = scan_elem

	# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	def setSearchContent(self, xml, recStatus):
		self.type = "search"
		self.content = xml
		self.recordStatus = recStatus

	# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	def getProtocolDocument(self):
		try:
			# locate major elements
			header_elem = self.doc.documentElement.firstChild
			diagnostics_elem = self.doc.getElementsByTagName("diagnostics")[0]
		
			# add response type
			if self.requestObj is not None and self.requestObj.type is not None :
				text = self.doc.createTextNode( self.requestObj.type )
				tmp_elem = header_elem.getElementsByTagName("type")[0]
				tmp_elem.appendChild(text)
			
			# add DBmod used:
			if self.dbmodVersion is not None:
				try:
					tmp_elem = header_elem.getElementsByTagName("version")[2]
					text = self.doc.createTextNode(self.dbmodVersion)
					tmp_elem.appendChild(text)
				except:
					pass
			# create content element
			tmp_elem = self.doc.createElementNS(self.__class__.__protocolNS__,"content")
			if self.recordStatus.total != None:
				tmp_elem.setAttribute('totalSearchHits', str(self.recordStatus.total))
			if self.recordStatus.drop != None:
				tmp_elem.setAttribute('recordDropped', str(self.recordStatus.drop))
			if self.recordStatus.count != None:
				tmp_elem.setAttribute('recordCount', str(self.recordStatus.count))
			if self.recordStatus.start != None:
				tmp_elem.setAttribute('recordStart', str(self.recordStatus.start))
			# has real content? no error ?
			if self.requestObj is not None and self.requestObj.type not in (None, "error") and isinstance(self.content, xml.dom.minidom.Node):
				tmp_elem.appendChild(self.content)
			# attach to document
			self.doc.documentElement.insertBefore(tmp_elem, diagnostics_elem)
			# add sendTime header info (2001-09-11T09:30:47-05:00)
			text = self.doc.createTextNode( convertDateTime2ISOString(now()) )
			tmp_elem = header_elem.getElementsByTagName("sendTime")[0]
			tmp_elem.appendChild(text)
		except:
			self.logObj + "Error while returning the XML document."

		# add processing time to log
		self.logObj.logPassedTime("create response")
		
		# add logging data to diagnostics
		for (log_type, err_code, logEntry) in self.logObj.returnLogList():
			tmp_elem = self.doc.createElementNS(self.__class__.__protocolNS__,"diagnostic")
			text = self.doc.createTextNode(logEntry)
			tmp_elem.appendChild(text)
			# add error code if existing
			if log_type is not None:
				tmp_elem.setAttribute('severity', str(log_type))
			if err_code is not None:
				tmp_elem.setAttribute('code', str(err_code))
			diagnostics_elem.appendChild(tmp_elem)
		#
		# return protocol document as DOM
		#
		return self.doc
		
	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def printResponseString(self, stream=sys.stdout, encoding='UTF-8', httpHeader=1):
		'''Returns a response as a string inserted into a stream.
		If parameter "stream" is None, the DOM object is returned. Otherwise the response is printed into the stream.
		The parameter "httpHeader" is used to include a http header.
		'''
		if httpHeader > 0:
			stream.write('Content-Type: text/xml\n\n') # Blank line marking end of HTTP headers
		# print / return document
		xml.dom.ext.PrettyPrint(self.getProtocolDocument(), stream=stream, encoding=encoding)
		self.writeLogs()		

	
	




