# -*- coding: utf8 -*-
'''
R E Q U E S T - C L A S S
**********************************************
*	   R E Q U E S T - C L A S S
*	   base class for request parsers
**********************************************
$RCSfile: base_request.py,v $
$Revision: 9 $
$Author: markus $
$Date: 2005-07-04 11:47:00 +0200 (Mon, 04 Jul 2005) $

Base class to generate a request object with the help of
a protocol specific request parser.
'''

import types
from xml.sax							import parse, parseString
from biocase.wrapper.errorclasses		import *
from biocase.wrapper.sql.operators		import compressOperator
from biocase.wrapper.protocol.BioCASe.biocase_handler import filterHandler as defaultFilterHandler


# ------------------------------------------------------------------------------------
class RequestBaseClass(object):
	'''Base request class. Use a protocol parser to fill it.'''
	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def __init__(self, logObj, psfObj):
		# store reference to log and psf objects
		self.logObj = logObj
		self.psfObj = psfObj
		# PROPERTIES
		self.doc = u''	        # is logged		# holds the request as a string if protocol is string based
		# protocol header & logging information
		self.type = None        # is logged
		self.sources = []       # is logged
		#
		# only for search requests:
		#---------------
		# info directly taken from request
		self.reqSchema = ''						# holds the name of the requested query-schema
		self.resSchema = ''						# holds the name of the requested response schema
		self.filter = None		# is logged		# holds WhereClause object of filter part
		self.count = 0			# is logged		# only count results ? default = false(0)
		self.start = 0			# is logged		# start record of output
		self.limit = 0			# is logged		# number of records returned ( 0 = unlimited)
		self.scan  = ''			# is logged		# xpath to element meant for scanning (scan type only)

	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def init(self, args):
		'''Override this method in the protocol specific request class to
		create the request object from passed parameters.'''
		# look for debugging parameters:
		self._checkForDebuggingArgs(args)
		# build request
		self.type = args.get('request_type', None)
		if self.type is not None:
			self.logObj + (2, "A " +self.type +" response requested.")
		else:
			self.logObj + "Unknown request type! Default to capabilities."
			self.type='capabilities'
		self.sources = args.get('sources', [])
		self.reqSchema = args.get('request_schema', None)
		self.resSchema = args.get('response_schema', None)
		self.filter = args.get('filter_obj', None)
		self.count = args.get('count', 0)
		self.start = args.get('start', 0)
		self.limit = args.get('limit', 0)
		self.scan  = args.get('concept_to_scan', None)

	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def setRequest(self, argHash=None, **args):
		'''Get a request, parse it & create a request object and then process it.'''
		if type(argHash) == types.DictType:
			args=argHash
		if len(args) < 1:
			# try to get the request from CGI environment
			args = self.getRequestHashFromCGI()
		# create request from request hash
		self.init(args)

	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def getRequestHashFromCGI(self):
		'''Override this method in the protocol specific request class to be able to get http based requests.'''
		raise "this protocol does not support http based requests."

	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def _checkForDebuggingArgs(self, args):
		'''Takes the request hash and checks it for existing debugging options.'''
		if args.has_key('debug'):
			try:
				debuglevel = int(args['debug'])
			except:
				self.logObj + "Invalid debug level found: %s" % str(args['debug'])
				return
			if debuglevel > 9:
				# for debugging levels above 10 take only the last digit and turn CMF pickling off !
				debuglevel = int(args['debug'][-1])
				self.psfObj.pickleCMF = 0
				self.logObj + "Pickling turned off - overriding provider setup"
			self.logObj + "Found online debuglevel setting: %i" % debuglevel
			self.logObj.setDebugLevel( debuglevel )
			del args['debug']

				
	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def areSchemasSupported(self, req=1, resp=0):
		''' check if requested schemas are supported.
		Set req or resp flags to 1 to check for request or response respectively.'''
		if req == 1:
			if self.psfObj.schemas.has_key(self.resSchema):
				if self.psfObj.schemas[self.resSchema].response > 0:
					return 1
			raise RequestedSchemaError("response", self.resSchema, self.logObj)
		if resp == 1:
			if not(self.psfObj.schemas.has_key(schemaname) and self.psfObj.schemas[schemaname].request > 0):
				raise RequestedSchemaError("request", schemaname, self.logObj)
		pass

	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def getFilterObjFromFilterString(self, CMFobj, filterString, filterHandler=defaultFilterHandler):
		''' Construct the SQL WHERE clause filter object from the protocol/filter string passed.
		Defaults to the biocase protocol filter handler.
		Needs a CMFobject of the requested namespace.
		Takes care of global CMF filters for each COP generated.'''
		self.logObj + (2, "Original filter string parsed: %s"%filterString)
		# build object with SAX Handler
		handler = filterHandler(CMFobj, self.logObj)
		parseString(filterString, handler)
		self.logObj + (2, "Filter created: %s" %unicode(handler.filterObj))
		self.filter = compressOperator( handler.filterObj )
		self.logObj + (1, "Compressed filter used: %s"%unicode(self.filter))
		#
		# update logging object with filter data:
		#
		self.logObj.setRequestData(self)
		return self.filter

	def getRequestDoc(self):
		return self.logObj.requestDoc
		
	def setRequestDoc(self, doc):
		self.logObj.requestDoc = doc
		
	# holds the request as a string if protocol is string based
	doc  = property( getRequestDoc, setRequestDoc )




