# -*- coding: utf8 -*-
'''
PSF MODULE
**********************
***  PSF MODULE
**********************
$RCSfile: psf.py,v $
$Revision: 575 $
$Author: markus $
$Date: 2006-03-16 16:58:44 +0100 (Thu, 16 Mar 2006) $

Functions associated in reading and parsing 
the system wide configuration file "ProviderConfiguration.xml".
'''

import os, os.path, xml.dom.ext, mx.DateTime
from stat import ST_MTIME
from xml.sax    import parse
from biocase.wrapper.errorclasses         import *
from biocase.wrapper.logging             import loggingClass
from biocase.wrapper.static_functions     import getText
from biocase.tools.various_functions     import unique, isTrue
from biocase.wrapper.psf_handler         import providerConfigHandler, SupportedSchema
from biocase.wrapper.graph.tablegraph     import tableGraphClass, tableAliasClass
import biocase.configuration

        

# ----------------------------------------------------------------------------
class ProviderSetup(object):
    PSFnamespace = ["http://www.biocase.org/schemas/providersetup/1.5"]

    def __init__(self, logObj=loggingClass()):
        # store reference to logging object
        self.logObj = logObj
        # get global project configurations
        self.cfg = biocase.configuration.Cfg('pywrapper')
        # read logging settings from cfg object
        self.logObj.logstate = self.cfg.pywrapper.logstate
        self.logObj.lograte  = self.cfg.pywrapper.lograte
        # the absolute filepath to relevant directories and files
        # they will be build by reading a specific database directory with method 'setFileDirPaths'
        self.configDir   = None
        self.PSF         = None
        #
        # defaults overridden in provider_setup_file.xml
        #
        self.sourcename    = None
        self.adminpassword = None
        self.recLimit      = self.getDefaultRecLimit()
        # host part
        self.user            = None
        self.password        = None
        self.datasource      = None
        self.dbIP            = None
        self.dbms            = None
        self.encoding        = None
        # schema part
        self.schemas      = {} # key=namespace, value = SupportedSchema Object
        # table part represented as an tablegraph object
        self.tablegraph     = tableGraphClass(logObj)

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def __repr__(self):
        return "PSF: sourcename=%s, configDir=%s, PSF=%s, IP=%s, recLimit=%s, pickleCMF=%s, adminpassword=%s, user=%s, password=%s, datasource=%s, dbIP=%s, dbms=%s, encoding=%s, schemas=%s, tablegraph=%s" % (self.sourcename, self.configDir, self.PSF, self.IP, self.recLimit, self.pickleCMF, self.adminpassword, self.user, self.password, self.datasource, self.dbIP, self.dbms, self.encoding, unicode(self.schemas), self.tablegraph)

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def repr(self):
        from biocase.tools.showobject import view
        return view(self)    
    
    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def setSourcename(self, sourcename):
        # store sourcename in psf and log object
        self.sourcename = sourcename
        self.logObj.setSourcename(sourcename)
        
    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def setConfigDirs(self, PSFpath=None, logDir=None):
        '''sets the relevant absolute paths to dirs and files'''
        if PSFpath is not None:
            self.configDir   = os.path.split(PSFpath)[0]
            self.PSF         = PSFpath
            # store sourcename in psf and log object
            try:
                sourcename = os.path.split(os.path.split(PSFpath)[0])[1]
            except:
                sourcename = "???"
            self.setSourcename(sourcename)
            self.logObj + (1,r"set PSF=%s" %(self.PSF))
        if logDir is not None:
            self.logObj.logdir = logDir
            self.logObj + (1,r"set logDir=%s" %(self.logObj.logdir) )

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def readPSFile(self, absFilename=None):
        #
        # use SAX to parse the providersetup.xml file and fill this object
        if absFilename == None:
            absFilename = self.PSF
        else:
            self.PSF = absFilename
        handler = providerConfigHandler(self)
        parse(absFilename, handler)
        try:
            handler = providerConfigHandler(self)
            parse(absFilename, handler)
        except:
            raise ProviderSetupParseError(absFilename, self.logObj)

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def writePSFile(self, filename=None, namespace=None):
        '''Writes the PSF object to an xml file.
        If no namespace is given it uses by default the latest PSF version.'''
        if filename is None:
            filename = self.PSF
        PSFile = file(filename,'w')
        self.logObj + (1, "writing PSFile to %s" %(filename) )
        xml.dom.ext.PrettyPrint(self.createPSFDOM(version=namespace), stream=PSFile, encoding='utf-8')
        PSFile.close()
        
    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def cleanup(self):
        '''Cleans up inconsistent PSF data without raising errors - like table without primary keys.'''
        status = True
        # verify used schemas
        used_namespaces={}
        for ns, schema in self.schemas.items():
            # check if file exists
            if not os.path.isfile(schema.conceptMappingFile):
                del self.schemas[ns]
                status = False
            # check for duplicate namespaces
            if used_namespaces.has_key(ns):
                del self.schemas[ns]
                status = False
            else:
                used_namespaces[ns] = 1
        # very tablealiases and graph structure
        for table in self.tablegraph.getTableAliasObjects():
            if len(table.pk) < 1 or table.table is None or table.alias is None:
                self.tablegraph.delAlias(table.alias)
                status = False
            for fk in table.fks:
                if fk.target is None or len(fk.attrList) < 1 or self.tablegraph.getTableAliasObj(fk.target) is None:
                    table.fks.remove(fk)
                    status = False
        return status
        
    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def getCMFfilename(self, namespace):
        '''Returns the full CMF filename used for a given namespace.
        If no CMF file is found None is returned'''
        self.logObj + "Try to get CMF for namespace %s"%namespace
        if self.schemas.has_key(namespace):
            return self.schemas[namespace].conceptMappingFile
        else:
            if self.schemas.has_key(namespace.lower()):
                return self.schemas[namespace.lower()].conceptMappingFile
            else:
                self.logObj + "No CMF found! Available CMFs: %s" %unicode(self.schemas.keys())
                return None

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def getPSFUpdatedWhen(self):
        '''Returns the mx.DateTime of the last update of the PSF file or None is unknown.'''
        if os.path.isfile(self.PSF):
            return mx.DateTime.TimestampFromTicks( os.stat(self.PSF)[ST_MTIME] )
        else:
            return None

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def getTableAliasObj(self, alias):
        '''Returns the table alias object for a given alias or None.'''
        try:
            return self.tablegraph.getTableAliasObj(alias)
        except:
            return None

    # ------------------------------------    
    def getTableAliasList(self):
        '''Returns a list of all table alias objects or [].'''
        try:
            return self.tablegraph.getTableAliasList()
        except:
            return []
        
    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def getCompletePKDBAttrObjList(self):
        '''Returns a list of all DBAttrObjects used for the primary keys of all table aliases.'''
        return self.getPKDBAttrObjListOfDBAttrObjList( aliasList=self.getTableAliasList() )
        
    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def getPKDBAttrObjListOfAlias(self, alias):
        '''Returns a list of DBAttrObjects for the primary keys of a given table alias (as an object or string).
        Returns [] if alias is not existing.'''
        pks = self.tablegraph.getTablePKsDict()
        if isinstance(alias, tableAliasClass):
            alias = alias.alias
        try:
            return pks[alias]
        except:
            return []

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def getPKDBAttrObjListOfDBAttrObjList(self, dbaoList=[], aliasList=[]):
        '''Returns a distinct list of DBAttrObjects for the primary keys of a given list of db-attribute objects.
        Returns [] if no tablealias is existing.'''
        pks = []
        # get the list of all table aliases
        for DBAttrObj in dbaoList:
            aliasList.append(DBAttrObj.table)
        # make them unique
        aliasList = unique(aliasList)
        # get the list of all PK-DBAttrObjects
        for alias in aliasList:
            pks += self.getPKDBAttrObjListOfAlias(alias)
        return pks
        
    # ------------------------------------    
    # graph related methods
    # ------------------------------------    

    # ------------------------------------    
    def buildGraph(self, aliasObjList):
        '''Build a new graph from a list of table alias objects overwriting any existing graph.'''
        self.tablegraph.buildGraph(aliasObjList)

    # ------------------------------------    
    def clearGraph(self):
        '''Remove nonsense data from tablegraph.'''
        self.tablegraph.clearGraph()

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def circlesExist(self):
        '''Returns 1 if the tablegraph contains circular structures.'''
        return self.tablegraph.circlesExist()
        
    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def getTableTreeObj(self, rootTableAliasName, staticTableAliasList=[]):
        '''function to create and return a tabletree object with 
        - a given root table-alias name and 
        - an optional list of static table aliases.
        Throws an error if a circular graph was given or if unknown aliases were passed.
        '''
        return self.tablegraph.getTableTree(rootTableAliasName, staticTableAliasList)


    # ------------------------------------    
    # schema related methods
    # ------------------------------------    

    # ------------------------------------    
    def delAllSchemas(self):
        '''Removes all supprted schemas of this psf.'''
        self.schemas = {}

    # ------------------------------------    
    def addSchema(self, schemaNS, file, request=1, response=1):
        '''Adds a supported schema to the psf.'''
        ss = SupportedSchema()
        ss.request = request
        ss.response= response
        ss.name = schemaNS
        ss.conceptMappingFile = file
        if self.schemas.has_key(schemaNS) and self.schemas[schemaNS] is not None:
            raise MultipleCMFsWithSameNamespaceError(self.logObj)
        self.schemas[schemaNS] = ss


    # ------------------------------------    
    # write methods
    # ------------------------------------    

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def createPSFDOM(self, version=None):
        '''    Returns a minidom object that represents this psf object as a ProviderSetupFile.
        '''
        from biocase.tools.xmlwriter import XMLwriterClass
        if version == None:
            # by default use the latest version
            version = self.__class__.PSFnamespace[-1]
        if version == self.__class__.PSFnamespace[0]:
            self.XMLwriter = XMLwriterClass(self.__class__.PSFnamespace[0])
            PSFdom = self.createPSFDOM_1_5()
        else:
            raise "unknown PSF version. Failed to write PSF."
        # free memory
        self.XMLwriter = None
        return PSFdom
        
    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def createPSFDOM_1_5(self):
        '''Return a PSF dom object with namespace version 1.5'''
        psf_node = self.XMLwriter.createNode("Provider", None, {'adminpassword':self.adminpassword})
        # HOST
        host_node = self.XMLwriter.createNode("Host")
        psf_node.appendChild(host_node)
        self.XMLwriter.addNode(host_node, "User", self.user)
        self.XMLwriter.addNode(host_node, "Password", self.password)
        self.XMLwriter.addNode(host_node, "Datasource", self.datasource)
        self.XMLwriter.addNode(host_node, "DB_IP", self.dbIP)
        self.XMLwriter.addNode(host_node, "DBMS", self.dbms)
        # DB
        db_node = self.XMLwriter.createNode("DB")
        psf_node.appendChild(db_node)
        self.XMLwriter.addNode(db_node, "TextEncoding", self.encoding)
        # table aliases
        tables = self.getTableAliasList()
        for t in tables:
            ta_node = self.XMLwriter.createNode("TableAlias", None, {'tablename':t.table, 'alias':t.alias})
            db_node.appendChild(ta_node)
            # PKS
            pk_node = self.XMLwriter.createNode("PrimaryKey")
            ta_node.appendChild(pk_node)
            for pk in t.pk:
                self.XMLwriter.addNode(pk_node, 'Attribute', pk.attr, {'dbType':pk.type})
            # FKS
            for fkObj in t.fks:
                fk_node = self.XMLwriter.createNode("ForeignKey", None, {'target':fkObj.target})
                ta_node.appendChild(fk_node)
                for attrObj in fkObj.attrList:
                    self.XMLwriter.addNode(fk_node, 'Attribute', attrObj.attr, {'dbType':attrObj.type})
        # SCHEMAS
        schemas_node = self.XMLwriter.createNode("SupportedSchemas")
        psf_node.appendChild(schemas_node)
        for ss in self.schemas.values():
            ss_node = self.XMLwriter.createNode("Schema", os.path.split(ss.conceptMappingFile)[1], {'namespace':ss.name, 'request':ss.request, 'response':ss.response})
            schemas_node.appendChild(ss_node)
        # WRAPPER BEHAVIOR
        behavior_node = self.XMLwriter.createNode("WrapperBehavior")
        psf_node.appendChild(behavior_node)
        if self.recLimit is not None:
            self.XMLwriter.addNode(behavior_node, "RecordResultLimit", str(self.recLimit))
        
        return psf_node

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def getIP(self):
        return self.cfg.host.replace(self.cfg.host[:self.cfg.host.find('://')+3], '' )
    def getDefaultRecLimit(self):
        try:
            limit = int(self.cfgObj.pywrapper.recLimit)
        except:
            limit = 0
        if limit < 0:
            limit = 0
        return limit
    def getPickleCMF(self):
        return isTrue(self.cfg.pywrapper.pickleCMF)

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    IP           = property(getIP)
    pickleCMF    = property(getPickleCMF)
