# -*- coding: utf8 -*-
'''
The BioCASE PyWrapper class, a XML/CGI interface to heterogeneous databases

$RCSfile: pywrapper.py,v $
$Revision: 569 $
$Author: markus $
$Date: 2006-03-09 16:10:04 +0100 (Thu, 09 Mar 2006) $
****************************************

+++  BioCASE	
+++  PyWrapper
+++  XML/CGI database interface

This is the main class of the pywrapper for BioCASE.
It implements the BioCASE protocol for query and response, based on DiGIR.
'''

# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is biocase.wrapper.
#
# The Initial Developer of the Original Code is
# M.D�ring, Botanic Garden and Botanical Museum Berlin-Dahlem.
# Portions created by the Initial Developer are Copyright (C) 2003
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# ***** END LICENSE BLOCK *****


# import error messages
from biocase.wrapper.errorclasses import *
from biocase.wrapper.psf import ProviderSetup
from biocase.wrapper.logging import loggingClass
import os, sys, types



# *********************************************
# ***  M A I N   Class
# *********************************************


# ----------------------------------------------------------------------------
class pywrapper:

	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def __init__(self, protocol='plaintext', logdir=None):
		'''Create a pywrapper instance based on the passed protocol.'''
		# create the log object used everywhere else (not global, but passed)
		self.logObj         = loggingClass()
		self.logObj.logdir  = logdir
		self.psfObj       = None
		self.protocolObj  = None
		self.resultDOM    = None
		self.resultStatus = None
		#
		# create protocol object (for request & response) first:
		#
		self._setProtocol(protocol)
			
	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def __del__(self):
		# force memory release
		self.protocolObj  = None
		self.psfObj       = None
		self.resultDOM    = None
		self.resultStatus = None
		self.logObj       = None
		
	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def _setProtocol(self, protocol):
		'''Define the protocol to be used for responding to a request.
		Setting protocol to "None" uses no protocol at all.'''
		if protocol.lower()   == "biocase":
			# biocase 
			from biocase.wrapper.protocol.BioCASe.biocase_protocol import ProtocolClass
			self.protocolObj = ProtocolClass(self.logObj)
		elif protocol.lower() == "species2000":
			# species2000 
			from biocase.wrapper.protocol.species2000.sp2k_protocol import ProtocolClass
			self.protocolObj = ProtocolClass(self.logObj)
		else:
			raise UnknownProtocolError(protocol)
		
	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def readConfigFiles(self, PSFPath='provider_setup_file.xml', logDir='.'):
		'''Reads the PSF and WCF files and creates a PSF object.
		Also updates the response and log objects by creating a reference to this new psf object.'''
		#
		# read provider-setup configuration file (psf)
		#
		try:
			psfObj = ProviderSetup(self.logObj)
			self.psfObj = psfObj
			self.psfObj.setConfigDirs(PSFpath=PSFPath, logDir=logDir)
			try:
				self.psfObj.readWCFile()
			except:
				pass
			self.psfObj.readPSFile()
			# update protocol object to include a psf object reference
			self.protocolObj.setPsfObj(self.psfObj)
		except:
			raise ProviderSetupParseError("UNKNOWN", self.logObj)
		pass

	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def setDatasourceName(self, name):
		if self._isPsfObjExisting():
			self.psfObj.setSourcename(name)
		else:
			raise NotPossibleToCreateDatasourceError()
			
	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def setRequest(self, argHash=None, **args):
		'''Inits a request and processes it.
		If no parameter hash is passed to this method, 
		it tries to get parameters via the protocol objects from the CGI environment.'''
		if self._isPsfObjExisting():
			if type(argHash) == types.DictType:
				args=argHash
			self.protocolObj.setRequest(args)
		else:
			raise PSFmissingError()

	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def printResponseString(self, stream=sys.stdout, encoding='UTF-8', httpHeader=1):
		'''Returns a XML response string. 
		A true httpHeader parameter will include an XML http header.'''
		try:
			self.protocolObj.printResponseString(stream, encoding, httpHeader)
		except:
			# nothing works. produce error message as plain text and try to log errors:
			try:
				self.logObj.logErrors()
			except:
				pass
		
	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def getFullLog(self, httpHeader=0):
		'''Returns the full log as a string.'''
		log = ''
		if httpHeader > 0:
			log = "Content-Type: text/plain\n\n"
		debuglevel = self.logObj.debuglevel
		self.logObj.setDebugLevel = 4
		log += str(self.logObj)
		self.logObj.setDebugLevel = debuglevel
		return log

	# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
	def _isPsfObjExisting(self):
		'''Checks if the PSF configuration has been read and returns 1 for true.'''
		if self.psfObj is None:
			return 0
		elif self.psfObj.__class__ == ProviderSetup:
			return 1
		else:
			return 0
